##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from txaio import time_ns
from zlmdb import flatbuffers

from .gen.log import MasterNodeUsage as MasterNodeUsageGen


class _MasterNodeUsage(MasterNodeUsageGen.MasterNodeUsage):
    """
    Expand methods on the class code generated by flatc.

    FIXME: comes up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsMasterNodeUsage(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _MasterNodeUsage()
        x.Init(buf, n + offset)
        return x

    def MrealmIdAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def PubkeyAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def ClientIpAddressAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MeteringIdAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class MasterNodeUsage(object):
    """
    Persisted master node metering record database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        self._timestamp = None
        self._mrealm_id = None
        self._timestamp_from = None
        self._pubkey = None
        self._client_ip_address = None
        self._client_ip_version = None
        self._client_ip_port = None
        self._seq = None
        self._sent = None
        self._processed = None
        self._status = None
        self._status_message = None
        self._metering_id = None

        self._count = None
        self._total = None
        self._nodes = None

        self._controllers = None
        self._hostmonitors = None
        self._routers = None
        self._containers = None
        self._guests = None
        self._proxies = None
        self._marketmakers = None

        self._sessions = None

        self._msgs_call = None
        self._msgs_yield = None
        self._msgs_invocation = None
        self._msgs_result = None
        self._msgs_error = None
        self._msgs_publish = None
        self._msgs_published = None
        self._msgs_event = None
        self._msgs_register = None
        self._msgs_registered = None
        self._msgs_subscribe = None
        self._msgs_subscribed = None

    def marshal(self):
        """

        :return:
        """
        obj = {
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "mrealm_id": str(self.mrealm_id) if self.mrealm_id else None,
            "timestamp_from": int(self.timestamp_from) if self.timestamp_from else None,
            "pubkey": bytes(self.pubkey) if self.pubkey else None,
            "client_ip_address": bytes(self.client_ip_address) if self.client_ip_address else None,
            "client_ip_version": self.client_ip_version,
            "client_ip_port": self.client_ip_port,
            "seq": self.seq,
            "sent": int(self.sent) if self.sent else None,
            "processed": int(self.processed) if self.processed else None,
            "status": self.status,
            "status_message": self.status_message,
            "metering_id": str(self.metering_id) if self.metering_id else None,
            "count": self.count,
            "total": self.total,
            "nodes": self.nodes,
            "controllers": self.controllers,
            "hostmonitors": self.hostmonitors,
            "routers": self.routers,
            "containers": self.containers,
            "guests": self.guests,
            "proxies": self.proxies,
            "marketmakers": self.marketmakers,
            "sessions": self.sessions,
            "msgs_call": self.msgs_call,
            "msgs_yield": self.msgs_yield,
            "msgs_invocation": self.msgs_invocation,
            "msgs_result": self.msgs_result,
            "msgs_error": self.msgs_error,
            "msgs_publish": self.msgs_publish,
            "msgs_published": self.msgs_published,
            "msgs_event": self.msgs_event,
            "msgs_register": self.msgs_register,
            "msgs_registered": self.msgs_registered,
            "msgs_subscribe": self.msgs_subscribe,
            "msgs_subscribed": self.msgs_subscribed,
        }
        return obj

    @staticmethod
    def parse(data):
        """

        :param data:
        :return:
        """
        assert type(data) == dict, 'data parsed must have type dict, but was "{}"'.format(type(data))
        obj = MasterNodeUsage()

        timestamp = data.get("timestamp", None)
        assert timestamp is None or type(timestamp) == int, '"timestamp" must have type int, but was "{}"'.format(
            type(timestamp)
        )
        if timestamp is None:
            # set current time as default
            obj._timestamp = np.datetime64(time_ns(), "ns")
        else:
            # set the value contained in the parsed data
            obj._timestamp = np.datetime64(timestamp, "ns")

        timestamp_from = data.get("timestamp_from", None)
        assert timestamp_from is None or type(timestamp_from) == int, (
            '"timestamp_from" must have type int, but was "{}"'.format(type(timestamp_from))
        )
        obj._timestamp_from = np.datetime64(timestamp_from, "ns") if timestamp_from is not None else None

        mrealm_id = data.get("mrealm_id", None)
        assert mrealm_id is None or type(mrealm_id) == str, '"mrealm_id" must have type str, but was "{}"'.format(
            type(mrealm_id)
        )
        if mrealm_id:
            obj._mrealm_id = uuid.UUID(mrealm_id)

        metering_id = data.get("metering_id", None)
        assert metering_id is None or type(metering_id) == str, (
            '"metering_id" must have type str, but was "{}"'.format(type(metering_id))
        )
        if metering_id:
            obj._metering_id = uuid.UUID(metering_id)

        pubkey = data.get("pubkey", None)
        assert pubkey is None or type(pubkey) == bytes and len(pubkey) == 32, (
            '"pubkey" must have type bytes of length 32, but was "{}" of length {}'.format(
                type(pubkey), len(pubkey) if type(pubkey) == bytes else None
            )
        )
        obj._pubkey = pubkey

        client_ip_address = data.get("client_ip_address", None)
        assert client_ip_address is None or type(client_ip_address) == bytes and len(client_ip_address) in [4, 16], (
            '"client_ip_address" must have type bytes of length 4 or 16, but was "{}" of length {}'.format(
                type(client_ip_address), len(client_ip_address) if type(client_ip_address) == bytes else None
            )
        )
        obj._client_ip_address = client_ip_address

        client_ip_version = data.get("client_ip_version", None)
        assert (
            client_ip_version is None
            or client_ip_version == 0
            or (type(client_ip_version) == int and client_ip_version in [4, 6])
        ), '"client_ip_version" must have value [4, 6], but was "{}"'.format(client_ip_version)
        obj._client_ip_version = client_ip_version

        client_ip_port = data.get("client_ip_port", None)
        assert (
            client_ip_port is None
            or client_ip_port == 0
            or (type(client_ip_port) == int and client_ip_port in range(2**16))
        ), '"client_ip_port" must have value [0, 2**16[, but was "{}"'.format(client_ip_port)
        obj._client_ip_port = client_ip_port

        seq = data.get("seq", None)
        assert seq is None or type(seq) == int, '"seq" must have type int, but was "{}"'.format(type(seq))
        obj._seq = seq

        sent = data.get("sent", None)
        assert sent is None or type(sent) == int, '"sent" must have type int, but was "{}"'.format(type(sent))
        if sent is not None:
            obj._sent = np.datetime64(sent, "ns") if sent else None

        processed = data.get("processed", None)
        assert processed is None or type(processed) == int, '"processed" must have type int, but was "{}"'.format(
            type(processed)
        )
        obj._processed = np.datetime64(processed, "ns") if processed else None

        status = data.get("status", 0)
        assert status is None or (type(status) == int and status in range(4)), (
            '"status" must have type int, but was "{}"'.format(type(status))
        )
        obj._status = status

        status_message = data.get("status_message", None)
        assert status_message is None or type(status_message) == str, (
            '"status_message" must have type str, but was "{}"'.format(type(status_message))
        )
        obj._status_message = status_message

        # metering data:

        count = data.get("count", None)
        assert count is None or type(count) == int
        obj._count = count

        total = data.get("total", None)
        assert total is None or type(total) == int
        obj._total = total

        nodes = data.get("nodes", None)
        assert nodes is None or type(nodes) == int
        obj._nodes = nodes

        controllers = data.get("controllers", None)
        assert controllers is None or type(controllers) == int
        obj._controllers = controllers

        hostmonitors = data.get("hostmonitors", None)
        assert hostmonitors is None or type(hostmonitors) == int
        obj._hostmonitors = hostmonitors

        routers = data.get("routers", None)
        assert routers is None or type(routers) == int
        obj._routers = routers

        containers = data.get("containers", None)
        assert containers is None or type(containers) == int
        obj._containers = containers

        guests = data.get("guests", None)
        assert guests is None or type(guests) == int
        obj._guests = guests

        proxies = data.get("proxies", None)
        assert proxies is None or type(proxies) == int
        obj._proxies = proxies

        marketmakers = data.get("marketmakers", None)
        assert marketmakers is None or type(marketmakers) == int
        obj._marketmakers = marketmakers

        sessions = data.get("sessions", None)
        assert sessions is None or type(sessions) == int
        obj._sessions = sessions

        msgs_call = data.get("msgs_call", None)
        assert msgs_call is None or type(msgs_call) == int
        obj._msgs_call = msgs_call

        msgs_yield = data.get("msgs_yield", None)
        assert msgs_yield is None or type(msgs_yield) == int
        obj._msgs_yield = msgs_yield

        msgs_invocation = data.get("msgs_invocation", None)
        assert msgs_invocation is None or type(msgs_invocation) == int
        obj._msgs_invocation = msgs_invocation

        msgs_result = data.get("msgs_result", None)
        assert msgs_result is None or type(msgs_result) == int
        obj._msgs_result = msgs_result

        msgs_error = data.get("msgs_error", None)
        assert msgs_error is None or type(msgs_error) == int
        obj._msgs_error = msgs_error

        msgs_publish = data.get("msgs_publish", None)
        assert msgs_publish is None or type(msgs_publish) == int
        obj._msgs_publish = msgs_publish

        msgs_published = data.get("msgs_published", None)
        assert msgs_published is None or type(msgs_published) == int
        obj._msgs_published = msgs_published

        msgs_event = data.get("msgs_event", None)
        assert msgs_event is None or type(msgs_event) == int
        obj._msgs_event = msgs_event

        msgs_register = data.get("msgs_register", None)
        assert msgs_register is None or type(msgs_register) == int
        obj._msgs_register = msgs_register

        msgs_registered = data.get("msgs_registered", None)
        assert msgs_registered is None or type(msgs_registered) == int
        obj._msgs_registered = msgs_registered

        msgs_subscribe = data.get("msgs_subscribe", None)
        assert msgs_subscribe is None or type(msgs_subscribe) == int
        obj._msgs_subscribe = msgs_subscribe

        msgs_subscribed = data.get("msgs_subscribed", None)
        assert msgs_subscribed is None or type(msgs_subscribed) == int
        obj._msgs_subscribed = msgs_subscribed

        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self):
        """

        :return: Timestamp when usage was recorded (at the node of recording).
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        assert isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def timestamp_from(self):
        """

        :return:
        """
        if self._timestamp_from is None and self._from_fbs:
            self._timestamp_from = np.datetime64(self._from_fbs.TimestampFrom(), "ns")
        return self._timestamp_from

    @timestamp_from.setter
    def timestamp_from(self, value):
        assert isinstance(value, np.datetime64)
        self._timestamp_from = value

    @property
    def mrealm_id(self):
        """

        :return:
        """
        if self._mrealm_id is None and self._from_fbs:
            if self._from_fbs.MrealmIdLength():
                _mrealm_id = self._from_fbs.MrealmIdAsBytes()
                if _mrealm_id:
                    self._mrealm_id = uuid.UUID(bytes=bytes(_mrealm_id))
        return self._mrealm_id

    @mrealm_id.setter
    def mrealm_id(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._mrealm_id = value

    @property
    def metering_id(self):
        """

        :return:
        """
        if self._metering_id is None and self._from_fbs:
            if self._from_fbs.MeteringIdLength():
                _metering_id = self._from_fbs.MeteringIdAsBytes()
                if _metering_id:
                    self._metering_id = uuid.UUID(bytes=bytes(_metering_id))
        return self._metering_id

    @metering_id.setter
    def metering_id(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._metering_id = value

    @property
    def pubkey(self):
        """

        :return:
        """
        if self._pubkey is None and self._from_fbs:
            if self._from_fbs.PubkeyLength():
                self._pubkey = self._from_fbs.PubkeyAsBytes()
        return self._pubkey

    @pubkey.setter
    def pubkey(self, value):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._pubkey = value

    @property
    def client_ip_address(self):
        """

        :return:
        """
        if self._client_ip_address is None and self._from_fbs:
            if self._from_fbs.ClientIpAddressLength():
                self._client_ip_address = self._from_fbs.ClientIpAddressAsBytes()
        return self._client_ip_address

    @client_ip_address.setter
    def client_ip_address(self, value):
        assert value is None or (type(value) == bytes and len(value) in [4, 16])
        self._client_ip_address = value

    @property
    def client_ip_version(self):
        """

        :return:
        """
        if self._client_ip_version is None and self._from_fbs:
            self._client_ip_version = self._from_fbs.ClientIpVersion()
        return self._client_ip_version

    @client_ip_version.setter
    def client_ip_version(self, value):
        assert value is None or (type(value) == int and value in [4, 6])
        self._client_ip_version = value

    @property
    def client_ip_port(self):
        """

        :return:
        """
        if self._client_ip_port is None and self._from_fbs:
            self._client_ip_port = self._from_fbs.ClientIpPort()
        return self._client_ip_port

    @client_ip_port.setter
    def client_ip_port(self, value):
        assert value is None or (type(value) == int and value in range(2**16))
        self._client_ip_port = value

    @property
    def seq(self):
        """

        :return:
        """
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq

    @seq.setter
    def seq(self, value):
        assert value is None or type(value) == int
        self._seq = value

    @property
    def sent(self):
        """

        :return:
        """
        if self._sent is None and self._from_fbs:
            self._sent = np.datetime64(self._from_fbs.Sent(), "ns")
        return self._sent

    @sent.setter
    def sent(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._sent = value

    @property
    def processed(self):
        """

        :return:
        """
        if self._processed is None and self._from_fbs:
            self._processed = np.datetime64(self._from_fbs.Processed(), "ns")
        return self._processed

    @processed.setter
    def processed(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._processed = value

    @property
    def status(self):
        """

        :return:
        """
        if self._status is None and self._from_fbs:
            self._status = self._from_fbs.Status()
        return self._status

    @status.setter
    def status(self, value):
        assert value is None or (type(value) == int and value in range(4))
        self._status = value

    @property
    def status_message(self):
        """

        :return:
        """
        if self._status_message is None and self._from_fbs:
            status_message = self._from_fbs.StatusMessage()
            if status_message:
                self._status_message = status_message.decode("utf8")
        return self._status_message

    @status_message.setter
    def status_message(self, value):
        assert value is None or type(value) == str
        self._status_message = value

    @property
    def count(self):
        """

        :return:
        """
        if self._count is None and self._from_fbs:
            self._count = self._from_fbs.Count()
        return self._count

    @count.setter
    def count(self, value):
        assert value is None or type(value) == int
        self._count = value

    @property
    def total(self):
        """

        :return:
        """
        if self._total is None and self._from_fbs:
            self._total = self._from_fbs.Total()
        return self._total

    @total.setter
    def total(self, value):
        assert value is None or type(value) == int
        self._total = value

    @property
    def nodes(self):
        """

        :return:
        """
        if self._nodes is None and self._from_fbs:
            self._nodes = self._from_fbs.Nodes()
        return self._nodes

    @nodes.setter
    def nodes(self, value):
        assert value is None or type(value) == int
        self._nodes = value

    @property
    def controllers(self):
        """

        :return:
        """
        if self._controllers is None and self._from_fbs:
            self._controllers = self._from_fbs.Controllers()
        return self._controllers

    @controllers.setter
    def controllers(self, value):
        assert value is None or type(value) == int
        self._controllers = value

    @property
    def hostmonitors(self):
        """

        :return:
        """
        if self._hostmonitors is None and self._from_fbs:
            self._hostmonitors = self._from_fbs.Hostmonitors()
        return self._hostmonitors

    @hostmonitors.setter
    def hostmonitors(self, value):
        assert value is None or type(value) == int
        self._hostmonitors = value

    @property
    def routers(self):
        """

        :return:
        """
        if self._routers is None and self._from_fbs:
            self._routers = self._from_fbs.Routers()
        return self._routers

    @routers.setter
    def routers(self, value):
        assert value is None or type(value) == int
        self._routers = value

    @property
    def containers(self):
        """

        :return:
        """
        if self._containers is None and self._from_fbs:
            self._containers = self._from_fbs.Containers()
        return self._containers

    @containers.setter
    def containers(self, value):
        assert value is None or type(value) == int
        self._containers = value

    @property
    def guests(self):
        """

        :return:
        """
        if self._guests is None and self._from_fbs:
            self._guests = self._from_fbs.Guests()
        return self._guests

    @guests.setter
    def guests(self, value):
        assert value is None or type(value) == int
        self._guests = value

    @property
    def proxies(self):
        """

        :return:
        """
        if self._proxies is None and self._from_fbs:
            self._proxies = self._from_fbs.Proxies()
        return self._proxies

    @proxies.setter
    def proxies(self, value):
        assert value is None or type(value) == int
        self._proxies = value

    @property
    def marketmakers(self):
        """

        :return:
        """
        if self._marketmakers is None and self._from_fbs:
            self._marketmakers = self._from_fbs.Marketmakers()
        return self._marketmakers

    @marketmakers.setter
    def marketmakers(self, value):
        assert value is None or type(value) == int
        self._marketmakers = value

    @property
    def sessions(self):
        """

        :return:
        """
        if self._sessions is None and self._from_fbs:
            self._sessions = self._from_fbs.Sessions()
        return self._sessions

    @sessions.setter
    def sessions(self, value):
        assert value is None or type(value) == int
        self._sessions = value

    @property
    def msgs_call(self):
        """

        :return:
        """
        if self._msgs_call is None and self._from_fbs:
            self._msgs_call = self._from_fbs.MsgsCall()
        return self._msgs_call

    @msgs_call.setter
    def msgs_call(self, value):
        assert value is None or type(value) == int
        self._msgs_call = value

    @property
    def msgs_yield(self):
        """

        :return:
        """
        if self._msgs_yield is None and self._from_fbs:
            self._msgs_yield = self._from_fbs.MsgsYield()
        return self._msgs_yield

    @msgs_yield.setter
    def msgs_yield(self, value):
        assert value is None or type(value) == int
        self._msgs_yield = value

    @property
    def msgs_invocation(self):
        """

        :return:
        """
        if self._msgs_invocation is None and self._from_fbs:
            self._msgs_invocation = self._from_fbs.MsgsInvocation()
        return self._msgs_invocation

    @msgs_invocation.setter
    def msgs_invocation(self, value):
        assert value is None or type(value) == int
        self._msgs_invocation = value

    @property
    def msgs_result(self):
        """

        :return:
        """
        if self._msgs_result is None and self._from_fbs:
            self._msgs_result = self._from_fbs.MsgsResult()
        return self._msgs_result

    @msgs_result.setter
    def msgs_result(self, value):
        assert value is None or type(value) == int
        self._msgs_result = value

    @property
    def msgs_error(self):
        """

        :return:
        """
        if self._msgs_error is None and self._from_fbs:
            self._msgs_error = self._from_fbs.MsgsError()
        return self._msgs_error

    @msgs_error.setter
    def msgs_error(self, value):
        assert value is None or type(value) == int
        self._msgs_error = value

    @property
    def msgs_publish(self):
        """

        :return:
        """
        if self._msgs_publish is None and self._from_fbs:
            self._msgs_publish = self._from_fbs.MsgsPublish()
        return self._msgs_publish

    @msgs_publish.setter
    def msgs_publish(self, value):
        assert value is None or type(value) == int
        self._msgs_publish = value

    @property
    def msgs_published(self):
        """

        :return:
        """
        if self._msgs_published is None and self._from_fbs:
            self._msgs_published = self._from_fbs.MsgsPublished()
        return self._msgs_published

    @msgs_published.setter
    def msgs_published(self, value):
        assert value is None or type(value) == int
        self._msgs_published = value

    @property
    def msgs_event(self):
        """

        :return:
        """
        if self._msgs_event is None and self._from_fbs:
            self._msgs_event = self._from_fbs.MsgsEvent()
        return self._msgs_event

    @msgs_event.setter
    def msgs_event(self, value):
        assert value is None or type(value) == int
        self._msgs_event = value

    @property
    def msgs_register(self):
        """

        :return:
        """
        if self._msgs_register is None and self._from_fbs:
            self._msgs_register = self._from_fbs.MsgsRegister()
        return self._msgs_register

    @msgs_register.setter
    def msgs_register(self, value):
        assert value is None or type(value) == int
        self._msgs_register = value

    @property
    def msgs_registered(self):
        """

        :return:
        """
        if self._msgs_registered is None and self._from_fbs:
            self._msgs_registered = self._from_fbs.MsgsRegistered()
        return self._msgs_registered

    @msgs_registered.setter
    def msgs_registered(self, value):
        assert value is None or type(value) == int
        self._msgs_registered = value

    @property
    def msgs_subscribe(self):
        """

        :return:
        """
        if self._msgs_subscribe is None and self._from_fbs:
            self._msgs_subscribe = self._from_fbs.MsgsSubscribe()
        return self._msgs_subscribe

    @msgs_subscribe.setter
    def msgs_subscribe(self, value):
        assert value is None or type(value) == int
        self._msgs_subscribe = value

    @property
    def msgs_subscribed(self):
        """

        :return:
        """
        if self._msgs_subscribed is None and self._from_fbs:
            self._msgs_subscribed = self._from_fbs.MsgsSubscribed()
        return self._msgs_subscribed

    @msgs_subscribed.setter
    def msgs_subscribed(self, value):
        assert value is None or type(value) == int
        self._msgs_subscribed = value

    @staticmethod
    def cast(buf):
        """

        :param buf:
        :return:
        """
        return MasterNodeUsage(_MasterNodeUsage.GetRootAsMasterNodeUsage(buf, 0))

    def build(self, builder):
        """

        :param builder:
        :return:
        """

        mrealm_id = self.mrealm_id.bytes if self.mrealm_id else None
        if mrealm_id:
            mrealm_id = builder.CreateString(mrealm_id)

        pubkey = self.pubkey
        if pubkey:
            pubkey = builder.CreateString(bytes(pubkey))

        client_ip_address = self.client_ip_address
        if client_ip_address:
            client_ip_address = builder.CreateString(bytes(client_ip_address))

        status_message = self.status_message
        if status_message:
            status_message = builder.CreateString(status_message)

        metering_id = self.metering_id.bytes if self.metering_id else None
        if metering_id:
            metering_id = builder.CreateString(metering_id)

        MasterNodeUsageGen.MasterNodeUsageStart(builder)

        if self.timestamp is not None:
            MasterNodeUsageGen.MasterNodeUsageAddTimestamp(builder, self.timestamp.astype("long"))

        if mrealm_id:
            MasterNodeUsageGen.MasterNodeUsageAddMrealmId(builder, mrealm_id)

        if self.timestamp_from is not None:
            MasterNodeUsageGen.MasterNodeUsageAddTimestampFrom(builder, self.timestamp_from.astype("long"))

        if pubkey:
            MasterNodeUsageGen.MasterNodeUsageAddPubkey(builder, pubkey)

        if client_ip_address:
            MasterNodeUsageGen.MasterNodeUsageAddClientIpAddress(builder, client_ip_address)

        if self.client_ip_version:
            MasterNodeUsageGen.MasterNodeUsageAddClientIpVersion(builder, self.client_ip_version)

        if self.client_ip_port:
            MasterNodeUsageGen.MasterNodeUsageAddClientIpPort(builder, self.client_ip_port)

        if self.seq:
            MasterNodeUsageGen.MasterNodeUsageAddSeq(builder, self.seq)

        if self.sent is not None:
            MasterNodeUsageGen.MasterNodeUsageAddSent(builder, self.sent.astype("long"))

        if self.processed is not None:
            MasterNodeUsageGen.MasterNodeUsageAddProcessed(builder, self.processed.astype("long"))

        if self.status:
            MasterNodeUsageGen.MasterNodeUsageAddStatus(builder, self.status)

        if status_message:
            MasterNodeUsageGen.MasterNodeUsageAddStatusMessage(builder, status_message)

        if metering_id:
            MasterNodeUsageGen.MasterNodeUsageAddMeteringId(builder, metering_id)

        if self.count:
            MasterNodeUsageGen.MasterNodeUsageAddCount(builder, self.count)

        if self.total:
            MasterNodeUsageGen.MasterNodeUsageAddTotal(builder, self.total)

        if self.nodes:
            MasterNodeUsageGen.MasterNodeUsageAddNodes(builder, self.nodes)

        if self.controllers:
            MasterNodeUsageGen.MasterNodeUsageAddControllers(builder, self.controllers)

        if self.hostmonitors:
            MasterNodeUsageGen.MasterNodeUsageAddHostmonitors(builder, self.hostmonitors)

        if self.routers:
            MasterNodeUsageGen.MasterNodeUsageAddRouters(builder, self.routers)

        if self.containers:
            MasterNodeUsageGen.MasterNodeUsageAddContainers(builder, self.containers)

        if self.guests:
            MasterNodeUsageGen.MasterNodeUsageAddGuests(builder, self.guests)

        if self.proxies:
            MasterNodeUsageGen.MasterNodeUsageAddProxies(builder, self.proxies)

        if self.marketmakers:
            MasterNodeUsageGen.MasterNodeUsageAddMarketmakers(builder, self.marketmakers)

        if self.sessions:
            MasterNodeUsageGen.MasterNodeUsageAddSessions(builder, self.sessions)

        if self.msgs_call:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsCall(builder, self.msgs_call)

        if self.msgs_yield:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsYield(builder, self.msgs_yield)

        if self.msgs_invocation:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsInvocation(builder, self.msgs_invocation)

        if self.msgs_result:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsResult(builder, self.msgs_result)

        if self.msgs_error:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsError(builder, self.msgs_error)

        if self.msgs_publish:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsPublish(builder, self.msgs_publish)

        if self.msgs_published:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsPublished(builder, self.msgs_published)

        if self.msgs_event:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsEvent(builder, self.msgs_event)

        if self.msgs_register:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsRegister(builder, self.msgs_register)

        if self.msgs_registered:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsRegistered(builder, self.msgs_registered)

        if self.msgs_subscribe:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsSubscribe(builder, self.msgs_subscribe)

        if self.msgs_subscribed:
            MasterNodeUsageGen.MasterNodeUsageAddMsgsSubscribed(builder, self.msgs_subscribed)

        final = MasterNodeUsageGen.MasterNodeUsageEnd(builder)

        return final



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import txaio

txaio.use_twisted()  # noqa

import numpy as np
import pytest
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.xbrmm import Transaction


def fill_transaction(transaction):
    now = time_ns()
    transaction.tid = uuid.uuid4()
    transaction.created = np.datetime64(now, "ns")
    transaction.created_payment_channel_seq = random.randint(1, 1000)
    transaction.created_paying_channel_seq = random.randint(1, 1000)
    transaction.offer = uuid.uuid4()
    transaction.amount = random.randint(1, 2**256 - 1)
    transaction.payment_channel = uuid.uuid4()
    transaction.paying_channel = uuid.uuid4()
    transaction.state = random.randint(1, 3)
    transaction.completed = np.datetime64(now, "ns")
    transaction.completed_payment_channel_seq = random.randint(1, 1000)
    transaction.completed_paying_channel_seq = random.randint(1, 1000)
    transaction.key = uuid.uuid4()
    transaction.buyer_pubkey = os.urandom(32)
    transaction.payment_channel_after = random.randint(1, 2**256 - 1)
    transaction.paying_channel_after = random.randint(1, 2**256 - 1)
    transaction.payment_mm_sig = os.urandom(65)
    transaction.payment_del_sig = os.urandom(65)
    transaction.paying_mm_sig = os.urandom(65)
    transaction.paying_del_sig = os.urandom(65)


def fill_transaction_empty(transaction):
    transaction.tid = None
    transaction.created = None
    transaction.created_payment_channel_seq = None
    transaction.created_paying_channel_seq = None
    transaction.offer = None
    transaction.amount = None
    transaction.payment_channel = None
    transaction.paying_channel = None
    transaction.state = None
    transaction.completed = None
    transaction.completed_payment_channel_seq = None
    transaction.completed_paying_channel_seq = None
    transaction.key = None
    transaction.buyer_pubkey = None
    transaction.payment_channel_after = None
    transaction.paying_channel_after = None
    transaction.payment_mm_sig = None
    transaction.payment_del_sig = None
    transaction.paying_mm_sig = None
    transaction.paying_del_sig = None


@pytest.fixture(scope="function")
def transaction():
    _transaction = Transaction()
    fill_transaction(_transaction)
    return _transaction


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


def test_transaction_roundtrip(transaction, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = transaction.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 720

    # create python object from bytes (flatbuffes)
    _transaction = Transaction.cast(data)

    assert _transaction.tid == transaction.tid
    assert _transaction.created == transaction.created
    assert _transaction.created_payment_channel_seq == transaction.created_payment_channel_seq
    assert _transaction.created_paying_channel_seq == transaction.created_paying_channel_seq
    assert _transaction.offer == transaction.offer
    assert _transaction.amount == transaction.amount
    assert _transaction.payment_channel == transaction.payment_channel
    assert _transaction.paying_channel == transaction.paying_channel
    assert _transaction.state == transaction.state
    assert _transaction.completed == transaction.completed
    assert _transaction.completed_payment_channel_seq == transaction.completed_payment_channel_seq
    assert _transaction.completed_paying_channel_seq == transaction.completed_paying_channel_seq
    assert _transaction.key == transaction.key
    assert _transaction.buyer_pubkey == transaction.buyer_pubkey
    assert _transaction.payment_channel_after == transaction.payment_channel_after
    assert _transaction.paying_channel_after == transaction.paying_channel_after
    assert _transaction.payment_mm_sig == transaction.payment_mm_sig
    assert _transaction.payment_del_sig == transaction.payment_del_sig
    assert _transaction.paying_mm_sig == transaction.paying_mm_sig
    assert _transaction.paying_del_sig == transaction.paying_del_sig


def test_transaction_empty(builder):
    transaction1 = Transaction()
    fill_transaction_empty(transaction1)

    # serialize to bytes (flatbuffers) from python object
    obj = transaction1.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 12

    # create python object from bytes (flatbuffes)
    transaction2 = Transaction.cast(data)

    unix_zero = np.datetime64(0, "ns")

    assert transaction2.tid is None
    assert transaction2.created == unix_zero
    assert transaction2.created_payment_channel_seq == 0
    assert transaction2.created_paying_channel_seq == 0
    assert transaction2.offer is None
    assert transaction2.amount == 0
    assert transaction2.payment_channel is None
    assert transaction2.paying_channel is None
    assert transaction2.state == 0
    assert transaction2.completed == unix_zero
    assert transaction2.completed_payment_channel_seq == 0
    assert transaction2.completed_paying_channel_seq == 0
    assert transaction2.key is None
    assert transaction2.buyer_pubkey is None
    assert transaction2.payment_channel_after == 0
    assert transaction2.paying_channel_after == 0
    assert transaction2.payment_mm_sig is None
    assert transaction2.payment_del_sig is None
    assert transaction2.paying_mm_sig is None
    assert transaction2.paying_del_sig is None


def test_transaction_roundtrip_perf(transaction, builder):
    obj = transaction.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _transaction = Transaction.cast(data)
        if True:
            assert _transaction.tid == transaction.tid
            assert _transaction.created == transaction.created
            assert _transaction.created_payment_channel_seq == transaction.created_payment_channel_seq
            assert _transaction.created_paying_channel_seq == transaction.created_paying_channel_seq
            assert _transaction.offer == transaction.offer
            assert _transaction.amount == transaction.amount
            assert _transaction.payment_channel == transaction.payment_channel
            assert _transaction.paying_channel == transaction.paying_channel
            assert _transaction.state == transaction.state
            assert _transaction.completed == transaction.completed
            assert _transaction.completed_payment_channel_seq == transaction.completed_payment_channel_seq
            assert _transaction.completed_paying_channel_seq == transaction.completed_paying_channel_seq
            assert _transaction.key == transaction.key
            assert _transaction.buyer_pubkey == transaction.buyer_pubkey
            assert _transaction.payment_channel_after == transaction.payment_channel_after
            assert _transaction.paying_channel_after == transaction.paying_channel_after
            assert _transaction.payment_mm_sig == transaction.payment_mm_sig
            assert _transaction.payment_del_sig == transaction.payment_del_sig
            assert _transaction.paying_mm_sig == transaction.paying_mm_sig
            assert _transaction.paying_del_sig == transaction.paying_del_sig

            scratch["value"] += _transaction.amount

    N = 7
    M = 20000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
from typing import Optional
from uuid import UUID


class ClusterNodeMembership(object):
    """
    Membership of a node in a cluster.
    """

    def __init__(self, cluster_oid: Optional[UUID] = None, node_oid: Optional[UUID] = None, _unknown=None):
        """

        :param cluster_oid: Object ID of the cluster the node is member in.

        :param node_oid: Object ID of the node that is member in the cluster.
        """
        self.cluster_oid = cluster_oid
        self.node_oid = node_oid
        self._unknown = _unknown

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        if other.cluster_oid != self.cluster_oid:
            return False
        if other.node_oid != self.node_oid:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return pprint.pformat(self.marshal())

    def marshal(self):
        """
        Marshal this object to a generic host language object.

        :return: dict
        """
        obj = {
            "cluster_oid": str(self.cluster_oid),
            "node_oid": str(self.node_oid),
        }
        return obj

    @staticmethod
    def parse(data):
        """
        Parse generic host language object into an object of this class.

        :param data: Generic host language object
        :type data: dict

        :return: instance of :class:`WebService`
        """
        assert type(data) == dict

        # future attributes (yet unknown) are not only ignored, but passed through!
        _unknown = {}
        for k in data:
            if k not in ["cluster_oid", "node_oid"]:
                _unknown[k] = data[k]

        cluster_oid = None
        if "cluster_oid" in data:
            assert type(data["cluster_oid"]) == str
            cluster_oid = UUID(data["cluster_oid"])

        node_oid = None
        if "node_oid" in data:
            assert type(data["node_oid"]) == str
            node_oid = UUID(data["node_oid"])

        obj = ClusterNodeMembership(cluster_oid=cluster_oid, node_oid=node_oid, _unknown=_unknown)

        return obj



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################
import pprint
import uuid

import numpy as np
from zlmdb import MapBytes20TimestampUuid, MapUuidFlatBuffers, flatbuffers, table

from cfxdb.gen.xbr import Catalog as CatalogGen


class _CatalogGen(CatalogGen.Catalog):
    @classmethod
    def GetRootAsCatalog(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _CatalogGen()
        x.Init(buf, n + offset)
        return x

    def OidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def CreatedAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def OwnerAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def TidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def SignatureAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Catalog(object):
    """
    ``XBRNetwork.FbsRepository`` database object.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (uuid);
        self._oid = None

        # uint64 (timestamp)
        self._timestamp = None

        # uint32
        self._seq = None

        # [uint8] (address)
        self._owner = None

        # string (multihash)
        self._terms = None

        # string (multihash)
        self._meta = None

        # [uint8] (ethhash)
        self._tid = None

        # [uint8] (ethsig)
        self._signature = None

    def marshal(self) -> dict:
        obj = {
            "oid": self.oid.bytes if self.oid else None,
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "seq": self.seq,
            "owner": bytes(self.owner) if self.owner else None,
            "terms": self.terms,
            "meta": self.meta,
            "tid": bytes(self.tid) if self.tid else None,
            "signature": bytes(self.signature) if self.signature else None,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def oid(self) -> uuid.UUID:
        """
        The unique ID of the market.
        """
        if self._oid is None and self._from_fbs:
            if self._from_fbs.OidLength():
                _oid = self._from_fbs.OidAsBytes()
                self._oid = uuid.UUID(bytes=bytes(_oid))
        return self._oid

    @oid.setter
    def oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._oid = value

    @property
    def timestamp(self) -> np.datetime64:
        """
        Database transaction time (epoch time in ns) of insert or last update.
        """
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def seq(self) -> int:
        """
        Global market sequence number.
        """
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq or 0

    @seq.setter
    def seq(self, value: int):
        assert value is None or type(value) == int
        self._seq = value

    @property
    def owner(self) -> bytes:
        """
        FbsRepository owner.
        """
        if self._owner is None and self._from_fbs:
            if self._from_fbs.OwnerLength():
                self._owner = self._from_fbs.OwnerAsBytes()
        return self._owner

    @owner.setter
    def owner(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 20)
        self._owner = value

    @property
    def terms(self) -> str:
        """
        The XBR market terms set by the market owner. IPFS Multihash pointing to a ZIP archive file with market documents.
        """
        if self._terms is None and self._from_fbs:
            terms = self._from_fbs.Terms()
            if terms:
                self._terms = terms.decode("utf8")
        return self._terms

    @terms.setter
    def terms(self, value: str):
        assert value is None or type(value) == str
        self._terms = value

    @property
    def meta(self) -> str:
        """
        The XBR market metadata published by the market owner. IPFS Multihash pointing to a RDF/Turtle file with market metadata.
        """
        if self._meta is None and self._from_fbs:
            meta = self._from_fbs.Meta()
            if meta:
                self._meta = meta.decode("utf8")
        return self._meta

    @meta.setter
    def meta(self, value):
        assert value is None or type(value) == str
        self._meta = value

    @property
    def tid(self) -> bytes:
        """
        Transaction hash of the transaction this change was committed to the blockchain under.
        """
        if self._tid is None and self._from_fbs:
            if self._from_fbs.TidLength():
                self._tid = self._from_fbs.TidAsBytes()
        return self._tid

    @tid.setter
    def tid(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 32)
        self._tid = value

    @property
    def signature(self) -> bytes:
        """
        When signed off-chain and submitted via ``XBRMarket.createMarketFor``.
        """
        if self._signature is None and self._from_fbs:
            if self._from_fbs.SignatureLength():
                self._signature = self._from_fbs.SignatureAsBytes()
        return self._signature

    @signature.setter
    def signature(self, value: bytes):
        assert value is None or (type(value) == bytes and len(value) == 65)
        self._signature = value

    @staticmethod
    def cast(buf):
        return Catalog(_CatalogGen.GetRootAsCatalog(buf, 0))

    def build(self, builder):
        oid = self.oid.bytes if self.oid else None
        if oid:
            oid = builder.CreateString(oid)

        terms = self.terms
        if terms:
            terms = builder.CreateString(terms)

        meta = self.meta
        if meta:
            meta = builder.CreateString(meta)

        tid = self.tid
        if tid:
            tid = builder.CreateString(tid)

        signature = self.signature
        if signature:
            signature = builder.CreateString(signature)

        owner = self.owner
        if owner:
            owner = builder.CreateString(owner)

        CatalogGen.CatalogStart(builder)

        if oid:
            CatalogGen.CatalogAddOid(builder, oid)

        if self.timestamp:
            CatalogGen.CatalogAddTimestamp(builder, int(self.timestamp))

        if self.seq:
            CatalogGen.CatalogAddSeq(builder, self.seq)

        if terms:
            CatalogGen.CatalogAddTerms(builder, terms)

        if meta:
            CatalogGen.CatalogAddMeta(builder, meta)

        if tid:
            CatalogGen.CatalogAddTid(builder, tid)

        if signature:
            CatalogGen.CatalogAddSignature(builder, signature)

        if owner:
            CatalogGen.CatalogAddOwner(builder, owner)

        final = CatalogGen.CatalogEnd(builder)

        return final


@table("60ba3189-d127-4522-bfbc-ed416bf7233c", build=Catalog.build, cast=Catalog.cast)
class Catalogs(MapUuidFlatBuffers):
    """
    Catalogs table, mapping from ``oid|UUID`` to :class:`cfxdb.xbr.FbsRepository`
    """


@table("2e331469-68f8-4b1b-a1b9-904f7e29dbc5")
class IndexCatalogsByOwner(MapBytes20TimestampUuid):
    """
    Catalogs-by-owner index with ``(owner_adr|bytes[20], created|int) -> catalog_id|UUID`` mapping.
    """



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import binascii
import os
import timeit
import uuid
from datetime import datetime

import cbor2
import pytest
import txaio
from autobahn import util
from zlmdb import flatbuffers

from cfxdb.user import ActivationToken, ActivationTokenFbs, User, UserFbs

txaio.use_twisted()


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


#
# ACTIVATION TOKEN
#


def fill_token(token):
    token.oid = uuid.uuid4()

    token.atype = 1
    token.status = 1
    token.created = datetime.utcnow()
    token.completed = None
    token.code = util.generate_activation_code()
    token.email = "homer.simpson@example.com"
    token.pubkey = binascii.b2a_hex(os.urandom(32)).decode()


@pytest.fixture(scope="function")
def token_fbs():
    _token = ActivationTokenFbs()
    fill_token(_token)
    return _token


@pytest.fixture(scope="function")
def token_cbor():
    _token = ActivationToken()
    fill_token(_token)
    return _token


def test_token_fbs_roundtrip_perf(token_fbs, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = token_fbs.build(builder)
    builder.Finish(obj)
    data = builder.Output()

    # create python object from bytes (flatbuffes)
    def loop():
        _token = ActivationTokenFbs.cast(data)
        if True:
            assert _token.oid == token_fbs.oid
            assert len(str(token_fbs.oid)) == 36
            assert _token.atype == token_fbs.atype
            assert _token.status == token_fbs.status
            assert _token.completed == token_fbs.completed
            assert _token.code == token_fbs.code
            assert _token.email == token_fbs.email
            assert _token.pubkey == token_fbs.pubkey
            assert len(token_fbs.pubkey) == 64

    N = 5
    M = 50000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000


def test_token_cbor_roundtrip_perf(token_cbor):
    # serialize to bytes (cbor) from python object
    obj = token_cbor.marshal()
    data = cbor2.dumps(obj)

    # create python object from bytes (cbor)
    def loop():
        _obj = cbor2.loads(data)
        _token = ActivationToken.parse(_obj)
        if True:
            assert _token.oid == token_cbor.oid
            assert _token.atype == token_cbor.atype
            assert _token.status == token_cbor.status
            assert _token.completed == token_cbor.completed
            assert _token.code == token_cbor.code
            assert _token.email == token_cbor.email
            assert _token.pubkey == token_cbor.pubkey

    N = 5
    M = 50000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000


#
# USER
#


def fill_user(user):
    user.oid = uuid.uuid4()

    user.label = "Homer the 3rd"
    user.description = "My motto as a user is: never read the f** manual;)"
    user.tags = ["geek", "pythonista", "lemon"]

    user.email = "homer.simpson@example.com"
    user.registered = datetime.utcnow()
    user.pubkey = binascii.b2a_hex(os.urandom(32)).decode()


@pytest.fixture(scope="function")
def user_fbs():
    _user = UserFbs()
    fill_user(_user)
    return _user


@pytest.fixture(scope="function")
def user_cbor():
    _user = User()
    fill_user(_user)
    return _user


def test_user_fbs_roundtrip_perf(user_fbs, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = user_fbs.build(builder)
    builder.Finish(obj)
    data = builder.Output()

    # create python object from bytes (flatbuffes)
    def loop():
        _user = UserFbs.cast(data)
        if True:
            assert _user.oid == user_fbs.oid
            assert len(str(_user.oid)) == 36
            assert _user.label == user_fbs.label
            assert _user.description == user_fbs.description
            assert _user.tags == user_fbs.tags

            assert _user.email == user_fbs.email
            assert _user.registered == user_fbs.registered
            assert _user.pubkey == user_fbs.pubkey
            assert len(_user.pubkey) == 64

    N = 5
    M = 50000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000


def test_user_cbor_roundtrip_perf(user_cbor):
    # serialize to bytes (cbor) from python object
    obj = user_cbor.marshal()
    data = cbor2.dumps(obj)

    # create python object from bytes (cbor)
    def loop():
        _obj = cbor2.loads(data)
        _user = User.parse(_obj)
        if True:
            assert _user.oid == user_cbor.oid
            assert _user.label == user_cbor.label
            assert _user.description == user_cbor.description
            assert _user.tags == user_cbor.tags
            assert _user.email == user_cbor.email
            assert _user.registered == user_cbor.registered
            assert _user.pubkey == user_cbor.pubkey

    N = 5
    M = 50000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: arealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# WAMP application realms defined.
class ApplicationRealm(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ApplicationRealm()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsApplicationRealm(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ApplicationRealm
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this object.
    # ApplicationRealm
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Time when the object was created.
    # ApplicationRealm
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # Owner organization of this object.
    # ApplicationRealm
    def Owner(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Name of this application realm (must be unique within the management realm at any given point in time).
    # ApplicationRealm
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # When dispatching events to receivers, batch sending in chunks of this many events (not all at once).
    # ApplicationRealm
    def EventDispatchingChunkSize(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint32Flags, o + self._tab.Pos)
        return 0

    # URI check level to enforce by the router.
    # ApplicationRealm
    def UriCheck(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Int8Flags, o + self._tab.Pos)
        return 0

    # Enable the WAMP meta API on the application realm.
    # ApplicationRealm
    def EnableMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False

    # Bridge the WAMP meta API from the application realm to the management realm, so that it can be tapped into.
    # ApplicationRealm
    def BridgeMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False


def ApplicationRealmStart(builder):
    builder.StartObject(8)


def Start(builder):
    ApplicationRealmStart(builder)


def ApplicationRealmAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    ApplicationRealmAddOid(builder, oid)


def ApplicationRealmAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    ApplicationRealmAddCreated(builder, created)


def ApplicationRealmAddOwner(builder, owner):
    builder.PrependStructSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(owner), 0)


def AddOwner(builder, owner):
    ApplicationRealmAddOwner(builder, owner)


def ApplicationRealmAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    ApplicationRealmAddName(builder, name)


def ApplicationRealmAddEventDispatchingChunkSize(builder, eventDispatchingChunkSize):
    builder.PrependUint32Slot(4, eventDispatchingChunkSize, 0)


def AddEventDispatchingChunkSize(builder, eventDispatchingChunkSize):
    ApplicationRealmAddEventDispatchingChunkSize(builder, eventDispatchingChunkSize)


def ApplicationRealmAddUriCheck(builder, uriCheck):
    builder.PrependInt8Slot(5, uriCheck, 0)


def AddUriCheck(builder, uriCheck):
    ApplicationRealmAddUriCheck(builder, uriCheck)


def ApplicationRealmAddEnableMetaApi(builder, enableMetaApi):
    builder.PrependBoolSlot(6, enableMetaApi, 0)


def AddEnableMetaApi(builder, enableMetaApi):
    ApplicationRealmAddEnableMetaApi(builder, enableMetaApi)


def ApplicationRealmAddBridgeMetaApi(builder, bridgeMetaApi):
    builder.PrependBoolSlot(7, bridgeMetaApi, 0)


def AddBridgeMetaApi(builder, bridgeMetaApi):
    ApplicationRealmAddBridgeMetaApi(builder, bridgeMetaApi)


def ApplicationRealmEnd(builder):
    return builder.EndObject()


def End(builder):
    return ApplicationRealmEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class UserKey(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = UserKey()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsUserKey(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # UserKey
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # User key - a Ed25519 public key - for authenticating using WAMP-cryptosign.
    # UserKey
    def Pubkey(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # UserKey
    def PubkeyAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # UserKey
    def PubkeyLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # UserKey
    def PubkeyIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Timestamp (epoch time in ns) of initial creation of this record.
    # UserKey
    def Created(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint64Flags, o + self._tab.Pos)
        return 0

    # ID of user account this user key is owned by.
    # UserKey
    def Owner(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags, a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1)
            )
        return 0

    # UserKey
    def OwnerAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # UserKey
    def OwnerLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # UserKey
    def OwnerIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0


def UserKeyStart(builder):
    builder.StartObject(3)


def Start(builder):
    UserKeyStart(builder)


def UserKeyAddPubkey(builder, pubkey):
    builder.PrependUOffsetTRelativeSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(pubkey), 0)


def AddPubkey(builder, pubkey):
    UserKeyAddPubkey(builder, pubkey)


def UserKeyStartPubkeyVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPubkeyVector(builder, numElems: int) -> int:
    return UserKeyStartPubkeyVector(builder, numElems)


def UserKeyAddCreated(builder, created):
    builder.PrependUint64Slot(1, created, 0)


def AddCreated(builder, created):
    UserKeyAddCreated(builder, created)


def UserKeyAddOwner(builder, owner):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(owner), 0)


def AddOwner(builder, owner):
    UserKeyAddOwner(builder, owner)


def UserKeyStartOwnerVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartOwnerVector(builder, numElems: int) -> int:
    return UserKeyStartOwnerVector(builder, numElems)


def UserKeyEnd(builder):
    return builder.EndObject()


def End(builder):
    return UserKeyEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import random
import timeit

import pytest
import zlmdb
from autobahn import util
from txaio import (
    time_ns,
    with_twisted,  # noqa
)
from zlmdb import flatbuffers

from cfxdb.realmstore import Event

zlmdb.TABLES_BY_UUID = {}


def fill_event(event):
    event.timestamp = time_ns()
    event.subscription = util.id()
    event.publication = util.id()
    event.receiver = util.id()
    event.retained = random.choice([True, False])
    event.acknowledged_delivery = random.choice([True, False])


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


@pytest.fixture(scope="function")
def event():
    _event = Event()
    fill_event(_event)
    return _event


def test_event_roundtrip(event, builder):
    obj = event.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) in [64, 56]

    _event = Event.cast(data)

    assert _event.timestamp == event.timestamp
    assert _event.subscription == event.subscription
    assert _event.publication == event.publication
    assert _event.receiver == event.receiver
    assert _event.retained == event.retained
    assert _event.acknowledged_delivery == event.acknowledged_delivery


def test_event_roundtrip_perf(event, builder):
    obj = event.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"timestamp": 0}

    def loop():
        _event = Event.cast(data)
        if True:
            assert _event.timestamp == event.timestamp
            assert _event.subscription == event.subscription
            assert _event.publication == event.publication
            assert _event.receiver == event.receiver
            assert _event.retained == event.retained
            assert _event.acknowledged_delivery == event.acknowledged_delivery

            scratch["timestamp"] += event.timestamp

    N = 5
    M = 100000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    assert scratch["timestamp"] > 0



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class ClusterNodeMembership(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = ClusterNodeMembership()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsClusterNodeMembership(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # ClusterNodeMembership
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # OID of web cluster to which to the node is added. A cluster can have zero or more nodes added.
    # ClusterNodeMembership
    def ClusterOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # OID of the node to add to the cluster. A node can be added to more than one cluster.
    # ClusterNodeMembership
    def NodeOid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # The desired (maximum) parallelism (in terms of CPU usage, given as number of workers) that the node should receive if it is active (not a standby node) in this cluster.
    # ClusterNodeMembership
    def Parallel(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint16Flags, o + self._tab.Pos)
        return 0

    # Wheather this node acts as a standby node that only takes over work when an active node fails.
    # ClusterNodeMembership
    def Standby(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos))
        return False


def ClusterNodeMembershipStart(builder):
    builder.StartObject(4)


def Start(builder):
    ClusterNodeMembershipStart(builder)


def ClusterNodeMembershipAddClusterOid(builder, clusterOid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(clusterOid), 0)


def AddClusterOid(builder, clusterOid):
    ClusterNodeMembershipAddClusterOid(builder, clusterOid)


def ClusterNodeMembershipAddNodeOid(builder, nodeOid):
    builder.PrependStructSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(nodeOid), 0)


def AddNodeOid(builder, nodeOid):
    ClusterNodeMembershipAddNodeOid(builder, nodeOid)


def ClusterNodeMembershipAddParallel(builder, parallel):
    builder.PrependUint16Slot(2, parallel, 0)


def AddParallel(builder, parallel):
    ClusterNodeMembershipAddParallel(builder, parallel)


def ClusterNodeMembershipAddStandby(builder, standby):
    builder.PrependBoolSlot(3, standby, 0)


def AddStandby(builder, standby):
    ClusterNodeMembershipAddStandby(builder, standby)


def ClusterNodeMembershipEnd(builder):
    return builder.EndObject()


def End(builder):
    return ClusterNodeMembershipEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: user

# CFC user roles (built into code, unconfigurable)
class UserRole(object):
    NONE = 0
    OWNER = 1
    ADMIN = 2
    USER = 3
    GUEST = 4



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbr

# XBR Market Actor type.
class ActorType(object):
    # Unset
    NONE = 0
    # Actor is a XBR Provider.
    PROVIDER = 1
    # Actor is a XBR Consumer.
    CONSUMER = 2
    # Actor is both a XBR Provider and XBR Consumer.
    PROVIDER_CONSUMER = 3



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: mrealm

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class RouterCluster(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = RouterCluster()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsRouterCluster(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # RouterCluster
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # ID of this object.
    # RouterCluster
    def Oid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = o + self._tab.Pos
            from ..oid_t import oid_t

            obj = oid_t()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Label for this object (not interpreted by CFC).
    # RouterCluster
    def Label(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Description for this object (not interpreted by CFC).
    # RouterCluster
    def Description(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Tags on this object.
    # RouterCluster
    def Tags(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.String(a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 4))
        return ""

    # RouterCluster
    def TagsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # RouterCluster
    def TagsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Unique user assigned name.
    # RouterCluster
    def Name(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Current status of web cluster.
    # RouterCluster
    def Status(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0


def RouterClusterStart(builder):
    builder.StartObject(6)


def Start(builder):
    RouterClusterStart(builder)


def RouterClusterAddOid(builder, oid):
    builder.PrependStructSlot(0, flatbuffers.number_types.UOffsetTFlags.py_type(oid), 0)


def AddOid(builder, oid):
    RouterClusterAddOid(builder, oid)


def RouterClusterAddLabel(builder, label):
    builder.PrependUOffsetTRelativeSlot(1, flatbuffers.number_types.UOffsetTFlags.py_type(label), 0)


def AddLabel(builder, label):
    RouterClusterAddLabel(builder, label)


def RouterClusterAddDescription(builder, description):
    builder.PrependUOffsetTRelativeSlot(2, flatbuffers.number_types.UOffsetTFlags.py_type(description), 0)


def AddDescription(builder, description):
    RouterClusterAddDescription(builder, description)


def RouterClusterAddTags(builder, tags):
    builder.PrependUOffsetTRelativeSlot(3, flatbuffers.number_types.UOffsetTFlags.py_type(tags), 0)


def AddTags(builder, tags):
    RouterClusterAddTags(builder, tags)


def RouterClusterStartTagsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartTagsVector(builder, numElems: int) -> int:
    return RouterClusterStartTagsVector(builder, numElems)


def RouterClusterAddName(builder, name):
    builder.PrependUOffsetTRelativeSlot(4, flatbuffers.number_types.UOffsetTFlags.py_type(name), 0)


def AddName(builder, name):
    RouterClusterAddName(builder, name)


def RouterClusterAddStatus(builder, status):
    builder.PrependUint8Slot(5, status, 0)


def AddStatus(builder, status):
    RouterClusterAddStatus(builder, status)


def RouterClusterEnd(builder):
    return builder.EndObject()


def End(builder):
    return RouterClusterEnd(builder)



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import os
import random
import timeit
import uuid

import numpy as np
import pytest
import txaio
import zlmdb
from zlmdb import flatbuffers

txaio.use_twisted()  # noqa

from autobahn.util import generate_activation_code
from txaio import time_ns

from cfxdb.tests._util import _gen_ipfs_hash
from cfxdb.xbrnetwork import Account, UserKey, VerifiedAction

zlmdb.TABLES_BY_UUID = {}


@pytest.fixture(scope="function")
def builder():
    _builder = flatbuffers.Builder(0)
    return _builder


#
# Account
#


def fill_account(account):
    now = time_ns()
    account.oid = uuid.uuid4()
    account.created = np.datetime64(now, "ns")
    account.username = "user{}".format(random.randint(0, 1000))
    account.email = "{}@example.com".format(account.username)
    account.email_verified = np.datetime64(now + random.randint(10 * 10**9, 60 * 10**9), "ns")
    account.wallet_type = random.randint(1, 3)
    account.wallet_address = os.urandom(20)
    account.registered = random.randint(0, 2**256 - 1)
    account.eula = _gen_ipfs_hash()
    account.profile = _gen_ipfs_hash()
    account.level = random.randint(1, 5)


@pytest.fixture(scope="function")
def account():
    _account = Account()
    fill_account(_account)
    return _account


def test_account_roundtrip(account, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = account.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 328

    # create python object from bytes (flatbuffes)
    _account = Account.cast(data)

    assert _account.oid == account.oid
    assert _account.created == account.created
    assert _account.username == account.username
    assert _account.email == account.email
    assert _account.email_verified == account.email_verified
    assert _account.wallet_type == account.wallet_type
    assert _account.wallet_address == account.wallet_address
    assert _account.registered == account.registered
    assert _account.eula == account.eula
    assert _account.profile == account.profile
    assert _account.level == account.level


def test_account_roundtrip_perf(account, builder):
    obj = account.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _account = Account.cast(data)
        if True:
            assert _account.oid == account.oid
            assert _account.created == account.created
            assert _account.username == account.username
            assert _account.email == account.email
            assert _account.email_verified == account.email_verified
            assert _account.wallet_type == account.wallet_type
            assert _account.wallet_address == account.wallet_address
            assert _account.registered == account.registered
            assert _account.eula == account.eula
            assert _account.profile == account.profile
            assert _account.level == account.level

            scratch["value"] += _account.level

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])


#
# VerifiedAction
#


def fill_vaction(vaction):
    vaction.oid = uuid.uuid4()
    vaction.created = np.datetime64(time_ns(), "ns")
    vaction.vtype = random.randint(1, 4)
    vaction.vstatus = random.randint(1, 4)
    vaction.vcode = generate_activation_code()
    vaction.verified_oid = uuid.uuid4()
    vaction.verified_data = {
        "f1": os.urandom(32),
        "f2": random.randint(1, 100),
        "f3": list(range(10)),
        "f4": generate_activation_code(),
    }


@pytest.fixture(scope="function")
def vaction():
    _vaction = VerifiedAction()
    fill_vaction(_vaction)
    return _vaction


def test_vaction_roundtrip(vaction, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = vaction.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 208

    # create python object from bytes (flatbuffes)
    _vaction = VerifiedAction.cast(data)

    assert _vaction.oid == vaction.oid
    assert _vaction.created == vaction.created
    assert _vaction.vtype == vaction.vtype
    assert _vaction.vstatus == vaction.vstatus
    assert _vaction.vcode == vaction.vcode
    assert _vaction.verified_oid == vaction.verified_oid
    assert _vaction.verified_data == vaction.verified_data


def test_vaction_roundtrip_perf(vaction, builder):
    obj = vaction.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _vaction = VerifiedAction.cast(data)
        if True:
            assert _vaction.oid == vaction.oid
            assert _vaction.created == vaction.created
            assert _vaction.vtype == vaction.vtype
            assert _vaction.vstatus == vaction.vstatus
            assert _vaction.vcode == vaction.vcode
            assert _vaction.verified_oid == vaction.verified_oid
            assert _vaction.verified_data == vaction.verified_data

            scratch["value"] += _vaction.vstatus

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])


#
# UserKey
#


def fill_userkey(userkey):
    userkey.pubkey = os.urandom(32)
    userkey.created = np.datetime64(time_ns(), "ns")
    userkey.owner = uuid.uuid4()


@pytest.fixture(scope="function")
def userkey():
    _userkey = UserKey()
    fill_userkey(_userkey)
    return _userkey


def test_userkey_roundtrip(userkey, builder):
    # serialize to bytes (flatbuffers) from python object
    obj = userkey.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    assert len(data) == 104

    # create python object from bytes (flatbuffes)
    _userkey = UserKey.cast(data)

    assert _userkey.pubkey == userkey.pubkey
    assert _userkey.created == userkey.created
    assert _userkey.owner == userkey.owner


def test_userkey_roundtrip_perf(userkey, builder):
    obj = userkey.build(builder)
    builder.Finish(obj)
    data = builder.Output()
    scratch = {"value": 0}

    def loop():
        _userkey = UserKey.cast(data)
        if True:
            assert _userkey.pubkey == userkey.pubkey
            assert _userkey.created == userkey.created
            assert _userkey.owner == userkey.owner

            scratch["value"] += int(_userkey.created)

    N = 5
    M = 10000
    samples = []
    print("measuring:")
    for i in range(N):
        secs = timeit.timeit(loop, number=M)
        ops = round(float(M) / secs, 1)
        samples.append(ops)
        print("{} objects/sec performance".format(ops))

    samples = sorted(samples)
    ops50 = samples[int(len(samples) / 2)]
    print("RESULT: {} objects/sec median performance".format(ops50))

    assert ops50 > 1000
    print(scratch["value"])



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid

import numpy as np
from txaio import time_ns
from zlmdb import flatbuffers

from cfxdb.gen.log import MWorkerLog as MWorkerLogGen


class _MWorkerLogGen(MWorkerLogGen.MWorkerLog):
    """
    Expand methods on the class code generated by flatc.

    FIXME: come up with a PR for flatc to generated this stuff automatically.
    """

    @classmethod
    def GetRootAsMWorkerLog(cls, buf, offset):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _MWorkerLogGen()
        x.Init(buf, n + offset)
        return x

    def NodeIdAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def MrealmIdAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class MWorkerLog(object):
    WORKER_TYPES = {
        "": 0,
        "router": 1,
        "container": 2,
        "guest": 3,
        "testee": 4,
        "proxy": 5,
        "xbrmm": 6,
        "hostmonitor": 7,
        "controller": 8,
    }
    WORKER_TYPENAMES = {
        0: "",
        1: "router",
        2: "container",
        3: "guest",
        4: "testee",
        5: "proxy",
        6: "xbrmm",
        7: "hostmonitor",
        8: "controller",
    }

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # PK uint64
        self._timestamp = None

        # PK node_id: [uint8] (uuid);
        self._node_id = None

        # PK worker_id: string;
        self._worker_id = None

        # period: uint32;
        self._period = None

        # mrealm_id: [uint8] (uuid);
        self._mrealm_id = None

        # _type: MWorkerType;
        self._type = None

        # state: MWorkerState;
        self._state = None

        # sent: uint64;
        self._sent = None

        # seq: uint64;
        self._seq = None

        # num_fds: uint32;
        self._num_fds = None

        # num_threads: uint16;
        self._num_threads = None

        # num_ctx_switches_involuntary: uint64;
        self._num_ctx_switches_involuntary = None

        # num_ctx_switches_involuntary_per_sec: uint32;
        self._num_ctx_switches_involuntary_per_sec = None

        # num_ctx_switches_voluntary: uint64;
        self._num_ctx_switches_voluntary = None

        # num_ctx_switches_voluntary_per_sec: uint32;
        self._num_ctx_switches_voluntary_per_sec = None

        # cpu_system: float;
        self._cpu_system = None

        # cpu_system_per_sec: float;
        self._cpu_system_per_sec = None

        # cpu_wait: float;
        self._cpu_wait = None

        # cpu_wait_per_sec: float;
        self._cpu_wait_per_sec = None

        # cpu_user: float;
        self._cpu_user = None

        # cpu_user_per_sec: float;
        self._cpu_user_per_sec = None

        # read_bytes: uint64;
        self._read_bytes = None

        # read_bytes_per_sec: uint64;
        self._read_bytes_per_sec = None

        # read_ios: uint64;
        self._read_ios = None

        # read_ios_per_sec: uint64;
        self._read_ios_per_sec = None

        # write_bytes: uint64;
        self._write_bytes = None

        # write_bytes_per_sec: uint64;
        self._write_bytes_per_sec = None

        # write_ios: uint64;
        self._write_ios = None

        # write_ios_per_sec: uint64;
        self._write_ios_per_sec = None

        # sent_bytes: uint64;
        self._sent_bytes = None

        # sent_bytes_per_sec: uint64;
        self._sent_bytes_per_sec = None

        # sent_ios: uint64;
        self._sent_ios = None

        # sent_ios_per_sec: uint64;
        self._sent_ios_per_sec = None

        # recv_bytes: uint64;
        self._recv_bytes = None

        # recv_bytes_per_sec: uint64;
        self._recv_bytes_per_sec = None

        # recv_ios: uint64;
        self._recv_ios = None

        # recv_ios_per_sec: uint64;
        self._recv_ios_per_sec = None

        # router_roles: uint64;
        self._router_roles = None

        # router_sessions: uint64;
        self._router_sessions = None

        # recv_call: uint64;
        self._recv_call = None

        # recv_yield: uint64;
        self._recv_yield = None

        # sent_invocation: uint64;
        self._sent_invocation = None

        # sent_result: uint64;
        self._sent_result = None

        # recv_publish: uint64;
        self._recv_publish = None

        # sent_published: uint64;
        self._sent_published = None

        # sent_event: uint64;
        self._sent_event = None

        # recv_register: uint64;
        self._recv_register = None

        # sent_registered: uint64;
        self._sent_registered = None

        # recv_subscribe: uint64;
        self._recv_subscribe = None

        # sent_subscribed: uint64;
        self._sent_subscribed = None

    @staticmethod
    def parse(mrealm_id, node_id, worker_id, heartbeat):
        assert isinstance(mrealm_id, uuid.UUID)
        assert isinstance(node_id, uuid.UUID)
        assert type(worker_id) == str
        assert type(heartbeat) == dict
        assert "timestamp" in heartbeat and type(heartbeat["timestamp"]) == int
        assert "seq" in heartbeat and type(heartbeat["seq"]) == int
        assert "type" in heartbeat and type(heartbeat["type"]) == str

        obj = MWorkerLog()
        obj._timestamp = np.datetime64(time_ns(), "ns")
        obj._period = heartbeat.get("period", None)
        obj._mrealm_id = mrealm_id
        obj._node_id = node_id
        obj._worker_id = worker_id
        obj._sent = np.datetime64(heartbeat["timestamp"], "ns") if heartbeat.get("timestamp", None) else None
        obj._seq = heartbeat.get("seq", None)
        obj._type = MWorkerLog.WORKER_TYPES.get(heartbeat.get("type", None), None)
        obj._state = heartbeat.get("state", None)

        process = heartbeat.get("process", {})
        obj._num_fds = process.get("num_fds", 0)
        obj._num_threads = process.get("num_threads", 0)
        obj._num_ctx_switches_involuntary = process.get("num_ctx_switches_involuntary", 0)
        obj._num_ctx_switches_involuntary_per_sec = process.get("num_ctx_switches_involuntary_per_sec", 0)
        obj._num_ctx_switches_voluntary = process.get("num_ctx_switches_voluntary", 0)
        obj._num_ctx_switches_voluntary_per_sec = process.get("num_ctx_switches_voluntary_per_sec", 0)
        obj._cpu_system = process.get("cpu_system", 0)
        obj._cpu_system_per_sec = process.get("cpu_system_per_sec", 0)
        obj._cpu_wait = process.get("cpu_wait", 0)
        obj._cpu_wait_per_sec = process.get("cpu_wait_per_sec", 0)
        obj._cpu_user = process.get("cpu_user", 0)
        obj._cpu_user_per_sec = process.get("cpu_user_per_sec", 0)
        obj._read_bytes = process.get("read_bytes", 0)
        obj._read_bytes_per_sec = process.get("read_bytes_per_sec", 0)
        obj._read_ios = process.get("read_ios", 0)
        obj._read_ios_per_sec = process.get("read_ios_per_sec", 0)
        obj._write_bytes = process.get("write_bytes", 0)
        obj._write_bytes_per_sec = process.get("write_bytes_per_sec", 0)
        obj._write_bytes_per_sec = process.get("write_bytes_per_sec", 0)
        obj._write_ios_per_sec = process.get("write_ios_per_sec", 0)
        obj._sent_bytes = process.get("sent_bytes", 0)
        obj._sent_bytes_per_sec = process.get("sent_bytes_per_sec", 0)
        obj._sent_ios = process.get("sent_ios", 0)
        obj._sent_ios_per_sec = process.get("sent_ios_per_sec", 0)
        obj._recv_bytes = process.get("recv_bytes", 0)
        obj._recv_bytes_per_sec = process.get("recv_bytes_per_sec", 0)
        obj._recv_ios = process.get("recv_ios", 0)
        obj._recv_ios_per_sec = process.get("recv_ios_per_sec", 0)

        router = heartbeat.get("router", {})
        obj._router_roles = router.get("roles", 0)
        obj._router_sessions = router.get("sessions", 0)

        messages = router.get("messages", {})
        obj._recv_call = messages.get("call", 0)
        obj._recv_yield = messages.get("yield", 0)
        obj._sent_invocation = messages.get("invocation", 0)
        obj._sent_result = messages.get("result", 0)
        obj._recv_publish = messages.get("publish", 0)
        obj._sent_published = messages.get("published", 0)
        obj._sent_event = messages.get("event", 0)
        obj._recv_register = messages.get("register", 0)
        obj._sent_registered = messages.get("registered", 0)
        obj._recv_subscribe = messages.get("subscribe", 0)
        obj._sent_subscribed = messages.get("subscribed", 0)

        return obj

    def marshal(self):
        obj = {
            "timestamp": int(self.timestamp) if self.timestamp else None,
            "period": self.period,
            "mrealm_id": str(self.mrealm_id) if self.mrealm_id else None,
            "node_id": str(self.node_id) if self.node_id else None,
            "worker_id": self.worker_id,
            "sent": int(self.sent) if self.sent else None,
            "seq": self.seq,
            "type": MWorkerLog.WORKER_TYPENAMES.get(self.type, None),
            "state": self.state,
            "process": {
                "num_fds": self.num_fds,
                "num_threads": self.num_threads,
                "num_ctx_switches_involuntary": self.num_ctx_switches_involuntary,
                "num_ctx_switches_involuntary_per_sec": self.num_ctx_switches_involuntary_per_sec,
                "num_ctx_switches_voluntary": self.num_ctx_switches_voluntary,
                "num_ctx_switches_voluntary_per_sec": self.num_ctx_switches_voluntary_per_sec,
                "cpu_system": self.cpu_system,
                "cpu_system_per_sec": self.cpu_system_per_sec,
                "cpu_wait": self.cpu_wait,
                "cpu_wait_per_sec": self.cpu_wait_per_sec,
                "cpu_user": self.cpu_user,
                "cpu_user_per_sec": self.cpu_user_per_sec,
                "read_bytes": self.read_bytes,
                "read_bytes_per_sec": self.read_bytes_per_sec,
                "read_ios": self.read_ios,
                "read_ios_per_sec": self.read_ios_per_sec,
                "write_bytes": self.write_bytes,
                "write_bytes_per_sec": self.write_bytes_per_sec,
                "write_ios": self.write_ios,
                "write_ios_per_sec": self.write_ios_per_sec,
                "sent_bytes": self.sent_bytes,
                "sent_bytes_per_sec": self.sent_bytes_per_sec,
                "sent_ios": self.sent_ios,
                "sent_ios_per_sec": self.sent_ios_per_sec,
                "recv_bytes": self.recv_bytes,
                "recv_bytes_per_sec": self.recv_bytes_per_sec,
                "recv_ios": self.recv_ios,
                "recv_ios_per_sec": self.recv_ios_per_sec,
            },
        }
        if self.type == 1:
            obj["router"] = {
                "roles": self.router_roles,
                "sessions": self.router_sessions,
                "messages": {
                    "call": self.recv_call,
                    "yield": self.recv_yield,
                    "invocation": self.sent_invocation,
                    "result": self.sent_result,
                    "publish": self.recv_publish,
                    "published": self.sent_published,
                    "event": self.sent_event,
                    "register": self.recv_register,
                    "registered": self.sent_registered,
                    "subscribe": self.recv_subscribe,
                    "subscribed": self.sent_subscribed,
                },
            }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def timestamp(self):
        if self._timestamp is None and self._from_fbs:
            self._timestamp = np.datetime64(self._from_fbs.Timestamp(), "ns")
        return self._timestamp

    @timestamp.setter
    def timestamp(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._timestamp = value

    @property
    def mrealm_id(self):
        if self._mrealm_id is None and self._from_fbs:
            if self._from_fbs.NodeIdLength():
                _mrealm_id = self._from_fbs.MrealmIdAsBytes()
                self._mrealm_id = uuid.UUID(bytes=bytes(_mrealm_id))
        return self._mrealm_id

    @mrealm_id.setter
    def mrealm_id(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._mrealm_id = value

    @property
    def node_id(self):
        if self._node_id is None and self._from_fbs:
            if self._from_fbs.NodeIdLength():
                _node_id = self._from_fbs.NodeIdAsBytes()
                self._node_id = uuid.UUID(bytes=bytes(_node_id))
        return self._node_id

    @node_id.setter
    def node_id(self, value):
        assert value is None or isinstance(value, uuid.UUID)
        self._node_id = value

    @property
    def worker_id(self):
        if self._worker_id is None and self._from_fbs:
            _worker_id = self._from_fbs.WorkerId()
            if _worker_id:
                self._worker_id = _worker_id.decode("utf8")
        return self._worker_id

    @worker_id.setter
    def worker_id(self, value):
        assert value is None or type(value) == str
        self._worker_id = value

    @property
    def period(self):
        if self._period is None and self._from_fbs:
            self._period = self._from_fbs.Period()
        return self._period

    @period.setter
    def period(self, value):
        assert value is None or type(value) == int
        self._period = value

    @property
    def type(self):
        if self._type is None and self._from_fbs:
            self._type = self._from_fbs.Type()
        return self._type

    @type.setter
    def type(self, value):
        assert value is None or type(value) == int
        self._type = value

    @property
    def state(self):
        if self._state is None and self._from_fbs:
            self._state = self._from_fbs.State()
        return self._state

    @state.setter
    def state(self, value):
        assert value is None or type(value) == int
        self._state = value

    @property
    def sent(self):
        if self._sent is None and self._from_fbs:
            self._sent = np.datetime64(self._from_fbs.Sent(), "ns")
        return self._sent

    @sent.setter
    def sent(self, value):
        assert value is None or isinstance(value, np.datetime64)
        self._sent = value

    @property
    def seq(self):
        if self._seq is None and self._from_fbs:
            self._seq = self._from_fbs.Seq()
        return self._seq

    @seq.setter
    def seq(self, value):
        assert value is None or type(value) == int
        self._seq = value

    @property
    def num_fds(self):
        if self._num_fds is None and self._from_fbs:
            self._num_fds = self._from_fbs.NumFds()
        return self._num_fds

    @num_fds.setter
    def num_fds(self, value):
        assert value is None or type(value) == int
        self._num_fds = value

    @property
    def num_threads(self):
        if self._num_threads is None and self._from_fbs:
            self._num_threads = self._from_fbs.NumThreads()
        return self._num_threads

    @num_threads.setter
    def num_threads(self, value):
        assert value is None or type(value) == int
        self._num_threads = value

    @property
    def num_ctx_switches_involuntary(self):
        if self._num_ctx_switches_involuntary is None and self._from_fbs:
            self._num_ctx_switches_involuntary = self._from_fbs.NumCtxSwitchesInvoluntary()
        return self._num_ctx_switches_involuntary

    @num_ctx_switches_involuntary.setter
    def num_ctx_switches_involuntary(self, value):
        assert value is None or type(value) == int
        self._num_ctx_switches_involuntary = value

    @property
    def num_ctx_switches_involuntary_per_sec(self):
        if self._num_ctx_switches_involuntary_per_sec is None and self._from_fbs:
            self._num_ctx_switches_involuntary_per_sec = self._from_fbs.NumCtxSwitchesInvoluntaryPerSec()
        return self._num_ctx_switches_involuntary_per_sec

    @num_ctx_switches_involuntary_per_sec.setter
    def num_ctx_switches_involuntary_per_sec(self, value):
        assert value is None or type(value) == int
        self._num_ctx_switches_involuntary_per_sec = value

    @property
    def num_ctx_switches_voluntary(self):
        if self._num_ctx_switches_voluntary is None and self._from_fbs:
            self._num_ctx_switches_voluntary = self._from_fbs.NumCtxSwitchesVoluntary()
        return self._num_ctx_switches_voluntary

    @num_ctx_switches_voluntary.setter
    def num_ctx_switches_voluntary(self, value):
        assert value is None or type(value) == int
        self._num_ctx_switches_voluntary = value

    @property
    def num_ctx_switches_voluntary_per_sec(self):
        if self._num_ctx_switches_voluntary_per_sec is None and self._from_fbs:
            self._num_ctx_switches_voluntary_per_sec = self._from_fbs.NumCtxSwitchesVoluntaryPerSec()
        return self._num_ctx_switches_voluntary_per_sec

    @num_ctx_switches_voluntary_per_sec.setter
    def num_ctx_switches_voluntary_per_sec(self, value):
        assert value is None or type(value) == int
        self._num_ctx_switches_voluntary_per_sec = value

    @property
    def cpu_system(self):
        if self._cpu_system is None and self._from_fbs:
            self._cpu_system = self._from_fbs.CpuSystem()
        return self._cpu_system

    @cpu_system.setter
    def cpu_system(self, value):
        assert value is None or type(value) == float
        self._cpu_system = value

    @property
    def cpu_system_per_sec(self):
        if self._cpu_system_per_sec is None and self._from_fbs:
            self._cpu_system_per_sec = self._from_fbs.CpuSystemPerSec()
        return self._cpu_system_per_sec

    @cpu_system_per_sec.setter
    def cpu_system_per_sec(self, value):
        assert value is None or type(value) == float
        self._cpu_system_per_sec = value

    @property
    def cpu_wait(self):
        if self._cpu_wait is None and self._from_fbs:
            self._cpu_wait = self._from_fbs.CpuWait()
        return self._cpu_wait

    @cpu_wait.setter
    def cpu_wait(self, value):
        assert value is None or type(value) == float
        self._cpu_wait = value

    @property
    def cpu_wait_per_sec(self):
        if self._cpu_wait_per_sec is None and self._from_fbs:
            self._cpu_wait_per_sec = self._from_fbs.CpuWaitPerSec()
        return self._cpu_wait_per_sec

    @cpu_wait_per_sec.setter
    def cpu_wait_per_sec(self, value):
        assert value is None or type(value) == float
        self._cpu_wait_per_sec = value

    @property
    def cpu_user(self):
        if self._cpu_user is None and self._from_fbs:
            self._cpu_user = self._from_fbs.CpuUser()
        return self._cpu_user

    @cpu_user.setter
    def cpu_user(self, value):
        assert value is None or type(value) == float
        self._cpu_user = value

    @property
    def cpu_user_per_sec(self):
        if self._cpu_user_per_sec is None and self._from_fbs:
            self._cpu_user_per_sec = self._from_fbs.CpuUserPerSec()
        return self._cpu_user_per_sec

    @cpu_user_per_sec.setter
    def cpu_user_per_sec(self, value):
        assert value is None or type(value) == float
        self._cpu_user_per_sec = value

    @property
    def read_bytes(self):
        if self._read_bytes is None and self._from_fbs:
            self._read_bytes = self._from_fbs.ReadBytes()
        return self._read_bytes

    @read_bytes.setter
    def read_bytes(self, value):
        assert value is None or type(value) == int
        self._read_bytes = value

    @property
    def read_bytes_per_sec(self):
        if self._read_bytes_per_sec is None and self._from_fbs:
            self._read_bytes_per_sec = self._from_fbs.ReadBytesPerSec()
        return self._read_bytes_per_sec

    @read_bytes_per_sec.setter
    def read_bytes_per_sec(self, value):
        assert value is None or type(value) == int
        self._read_bytes_per_sec = value

    @property
    def read_ios(self):
        if self._read_ios is None and self._from_fbs:
            self._read_ios = self._from_fbs.ReadIos()
        return self._read_ios

    @read_ios.setter
    def read_ios(self, value):
        assert value is None or type(value) == int
        self._read_ios = value

    @property
    def read_ios_per_sec(self):
        if self._read_ios_per_sec is None and self._from_fbs:
            self._read_ios_per_sec = self._from_fbs.ReadIosPerSec()
        return self._read_ios_per_sec

    @read_ios_per_sec.setter
    def read_ios_per_sec(self, value):
        assert value is None or type(value) == int
        self._read_ios_per_sec = value

    @property
    def write_bytes(self):
        if self._write_bytes is None and self._from_fbs:
            self._write_bytes = self._from_fbs.WriteBytes()
        return self._write_bytes

    @write_bytes.setter
    def write_bytes(self, value):
        assert value is None or type(value) == int
        self._write_bytes = value

    @property
    def write_bytes_per_sec(self):
        if self._write_bytes_per_sec is None and self._from_fbs:
            self._write_bytes_per_sec = self._from_fbs.WriteBytesPerSec()
        return self._write_bytes_per_sec

    @write_bytes_per_sec.setter
    def write_bytes_per_sec(self, value):
        assert value is None or type(value) == int
        self._write_bytes_per_sec = value

    @property
    def write_ios(self):
        if self._write_ios is None and self._from_fbs:
            self._write_ios = self._from_fbs.WriteIos()
        return self._write_ios

    @write_ios.setter
    def write_ios(self, value):
        assert value is None or type(value) == int
        self._write_ios = value

    @property
    def write_ios_per_sec(self):
        if self._write_ios_per_sec is None and self._from_fbs:
            self._write_ios_per_sec = self._from_fbs.WriteIosPerSec()
        return self._write_ios_per_sec

    @write_ios_per_sec.setter
    def write_ios_per_sec(self, value):
        assert value is None or type(value) == int
        self._write_ios_per_sec = value

    @property
    def sent_bytes(self):
        if self._sent_bytes is None and self._from_fbs:
            self._sent_bytes = self._from_fbs.SentBytes()
        return self._sent_bytes

    @sent_bytes.setter
    def sent_bytes(self, value):
        assert value is None or type(value) == int
        self._sent_bytes = value

    @property
    def sent_bytes_per_sec(self):
        if self._sent_bytes_per_sec is None and self._from_fbs:
            self._sent_bytes_per_sec = self._from_fbs.SentBytesPerSec()
        return self._sent_bytes_per_sec

    @sent_bytes_per_sec.setter
    def sent_bytes_per_sec(self, value):
        assert value is None or type(value) == int
        self._sent_bytes_per_sec = value

    @property
    def sent_ios(self):
        if self._sent_ios is None and self._from_fbs:
            self._sent_ios = self._from_fbs.SentIos()
        return self._sent_ios

    @sent_ios.setter
    def sent_ios(self, value):
        assert value is None or type(value) == int
        self._sent_ios = value

    @property
    def sent_ios_per_sec(self):
        if self._sent_ios_per_sec is None and self._from_fbs:
            self._sent_ios_per_sec = self._from_fbs.SentIosPerSec()
        return self._sent_ios_per_sec

    @sent_ios_per_sec.setter
    def sent_ios_per_sec(self, value):
        assert value is None or type(value) == int
        self._sent_ios_per_sec = value

    @property
    def recv_bytes(self):
        if self._recv_bytes is None and self._from_fbs:
            self._recv_bytes = self._from_fbs.RecvBytes()
        return self._recv_bytes

    @recv_bytes.setter
    def recv_bytes(self, value):
        assert value is None or type(value) == int
        self._recv_bytes = value

    @property
    def recv_bytes_per_sec(self):
        if self._recv_bytes_per_sec is None and self._from_fbs:
            self._recv_bytes_per_sec = self._from_fbs.RecvBytesPerSec()
        return self._recv_bytes_per_sec

    @recv_bytes_per_sec.setter
    def recv_bytes_per_sec(self, value):
        assert value is None or type(value) == int
        self._recv_bytes_per_sec = value

    @property
    def recv_ios(self):
        if self._recv_ios is None and self._from_fbs:
            self._recv_ios = self._from_fbs.RecvIos()
        return self._recv_ios

    @recv_ios.setter
    def recv_ios(self, value):
        assert value is None or type(value) == int
        self._recv_ios = value

    @property
    def recv_ios_per_sec(self):
        if self._recv_ios_per_sec is None and self._from_fbs:
            self._recv_ios_per_sec = self._from_fbs.RecvIosPerSec()
        return self._recv_ios_per_sec

    @recv_ios_per_sec.setter
    def recv_ios_per_sec(self, value):
        assert value is None or type(value) == int
        self._recv_ios_per_sec = value

    @property
    def router_roles(self):
        if self._router_roles is None and self._from_fbs:
            self._router_roles = self._from_fbs.RouterRoles()
        return self._router_roles

    @router_roles.setter
    def router_roles(self, value):
        assert value is None or type(value) == int
        self._router_roles = value

    @property
    def router_sessions(self):
        if self._router_sessions is None and self._from_fbs:
            self._router_sessions = self._from_fbs.RouterSessions()
        return self._router_sessions

    @router_sessions.setter
    def router_sessions(self, value):
        assert value is None or type(value) == int
        self._router_sessions = value

    @property
    def recv_call(self):
        if self._recv_call is None and self._from_fbs:
            self._recv_call = self._from_fbs.RecvCall()
        return self._recv_call

    @recv_call.setter
    def recv_call(self, value):
        assert value is None or type(value) == int
        self._recv_call = value

    @property
    def recv_yield(self):
        if self._recv_yield is None and self._from_fbs:
            self._recv_yield = self._from_fbs.RecvYield()
        return self._recv_yield

    @recv_yield.setter
    def recv_yield(self, value):
        assert value is None or type(value) == int
        self._recv_yield = value

    @property
    def sent_invocation(self):
        if self._sent_invocation is None and self._from_fbs:
            self._sent_invocation = self._from_fbs.SentInvocation()
        return self._sent_invocation

    @sent_invocation.setter
    def sent_invocation(self, value):
        assert value is None or type(value) == int
        self._sent_invocation = value

    @property
    def sent_result(self):
        if self._sent_result is None and self._from_fbs:
            self._sent_result = self._from_fbs.SentResult()
        return self._sent_result

    @sent_result.setter
    def sent_result(self, value):
        assert value is None or type(value) == int
        self._sent_result = value

    @property
    def recv_publish(self):
        if self._recv_publish is None and self._from_fbs:
            self._recv_publish = self._from_fbs.RecvPublish()
        return self._recv_publish

    @recv_publish.setter
    def recv_publish(self, value):
        assert value is None or type(value) == int
        self._recv_publish = value

    @property
    def sent_published(self):
        if self._sent_published is None and self._from_fbs:
            self._sent_published = self._from_fbs.SentPublished()
        return self._sent_published

    @sent_published.setter
    def sent_published(self, value):
        assert value is None or type(value) == int
        self._sent_published = value

    @property
    def sent_event(self):
        if self._sent_event is None and self._from_fbs:
            self._sent_event = self._from_fbs.SentEvent()
        return self._sent_event

    @sent_event.setter
    def sent_event(self, value):
        assert value is None or type(value) == int
        self._sent_event = value

    @property
    def recv_register(self):
        if self._recv_register is None and self._from_fbs:
            self._recv_register = self._from_fbs.RecvRegister()
        return self._recv_register

    @recv_register.setter
    def recv_register(self, value):
        assert value is None or type(value) == int
        self._recv_register = value

    @property
    def sent_registered(self):
        if self._sent_registered is None and self._from_fbs:
            self._sent_registered = self._from_fbs.SentRegistered()
        return self._sent_registered

    @sent_registered.setter
    def sent_registered(self, value):
        assert value is None or type(value) == int
        self._sent_registered = value

    @property
    def recv_subscribe(self):
        if self._recv_subscribe is None and self._from_fbs:
            self._recv_subscribe = self._from_fbs.RecvSubscribe()
        return self._recv_subscribe

    @recv_subscribe.setter
    def recv_subscribe(self, value):
        assert value is None or type(value) == int
        self._recv_subscribe = value

    @property
    def sent_subscribed(self):
        if self._sent_subscribed is None and self._from_fbs:
            self._sent_subscribed = self._from_fbs.SentSubscribed()
        return self._sent_subscribed

    @sent_subscribed.setter
    def sent_subscribed(self, value):
        assert value is None or type(value) == int
        self._sent_subscribed = value

    @staticmethod
    def cast(buf):
        return MWorkerLog(_MWorkerLogGen.GetRootAsMWorkerLog(buf, 0))

    def build(self, builder):
        mrealm_id = self.mrealm_id.bytes if self.mrealm_id else None
        if mrealm_id:
            mrealm_id = builder.CreateString(mrealm_id)

        node_id = self.node_id.bytes if self.node_id else None
        if node_id:
            node_id = builder.CreateString(node_id)

        worker_id = None
        if self.worker_id:
            worker_id = builder.CreateString(self.worker_id)

        MWorkerLogGen.MWorkerLogStart(builder)

        if self.timestamp:
            MWorkerLogGen.MWorkerLogAddTimestamp(builder, int(self.timestamp))

        if mrealm_id:
            MWorkerLogGen.MWorkerLogAddMrealmId(builder, mrealm_id)

        if node_id:
            MWorkerLogGen.MWorkerLogAddNodeId(builder, node_id)

        if worker_id:
            MWorkerLogGen.MWorkerLogAddWorkerId(builder, worker_id)

        if self.period:
            MWorkerLogGen.MWorkerLogAddPeriod(builder, self.period)

        if self.type:
            MWorkerLogGen.MWorkerLogAddType(builder, self.type)

        if self.state:
            MWorkerLogGen.MWorkerLogAddState(builder, self.state)

        if self.sent:
            MWorkerLogGen.MWorkerLogAddSent(builder, int(self.sent))

        if self.seq:
            MWorkerLogGen.MWorkerLogAddSeq(builder, self.seq)

        if self.num_fds:
            MWorkerLogGen.MWorkerLogAddNumFds(builder, self.num_fds)

        if self.num_threads:
            MWorkerLogGen.MWorkerLogAddNumThreads(builder, self.num_threads)

        if self.num_ctx_switches_involuntary:
            MWorkerLogGen.MWorkerLogAddNumCtxSwitchesInvoluntary(builder, self.num_ctx_switches_involuntary)

        if self.num_ctx_switches_involuntary_per_sec:
            MWorkerLogGen.MWorkerLogAddNumCtxSwitchesInvoluntaryPerSec(
                builder, self.num_ctx_switches_involuntary_per_sec
            )

        if self.num_ctx_switches_voluntary:
            MWorkerLogGen.MWorkerLogAddNumCtxSwitchesVoluntary(builder, self.num_ctx_switches_voluntary)

        if self.num_ctx_switches_voluntary_per_sec:
            MWorkerLogGen.MWorkerLogAddNumCtxSwitchesVoluntaryPerSec(builder, self.num_ctx_switches_voluntary_per_sec)

        if self.cpu_system:
            MWorkerLogGen.MWorkerLogAddCpuSystem(builder, self.cpu_system)

        if self.cpu_system_per_sec:
            MWorkerLogGen.MWorkerLogAddCpuSystemPerSec(builder, self.cpu_system_per_sec)

        if self.cpu_wait:
            MWorkerLogGen.MWorkerLogAddCpuWait(builder, self.cpu_wait)

        if self.cpu_wait_per_sec:
            MWorkerLogGen.MWorkerLogAddCpuWaitPerSec(builder, self.cpu_wait_per_sec)

        if self.cpu_user:
            MWorkerLogGen.MWorkerLogAddCpuUser(builder, self.cpu_user)

        if self.cpu_user_per_sec:
            MWorkerLogGen.MWorkerLogAddCpuUserPerSec(builder, self.cpu_user_per_sec)

        if self.read_bytes:
            MWorkerLogGen.MWorkerLogAddReadBytes(builder, self.read_bytes)

        if self.read_bytes_per_sec:
            MWorkerLogGen.MWorkerLogAddReadBytesPerSec(builder, self.read_bytes_per_sec)

        if self.read_ios:
            MWorkerLogGen.MWorkerLogAddReadIos(builder, self.read_ios)

        if self.read_ios_per_sec:
            MWorkerLogGen.MWorkerLogAddReadIosPerSec(builder, self.read_ios_per_sec)

        if self.write_bytes:
            MWorkerLogGen.MWorkerLogAddWriteBytes(builder, self.write_bytes)

        if self.write_bytes_per_sec:
            MWorkerLogGen.MWorkerLogAddWriteBytesPerSec(builder, self.write_bytes_per_sec)

        if self.write_ios:
            MWorkerLogGen.MWorkerLogAddWriteIos(builder, self.write_ios)

        if self.write_ios_per_sec:
            MWorkerLogGen.MWorkerLogAddWriteIosPerSec(builder, self.write_ios_per_sec)

        if self.sent_bytes:
            MWorkerLogGen.MWorkerLogAddSentBytes(builder, self.sent_bytes)

        if self.sent_bytes_per_sec:
            MWorkerLogGen.MWorkerLogAddSentBytesPerSec(builder, self.sent_bytes_per_sec)

        if self.sent_ios:
            MWorkerLogGen.MWorkerLogAddSentIos(builder, self.sent_ios)

        if self.sent_ios_per_sec:
            MWorkerLogGen.MWorkerLogAddSentIosPerSec(builder, self.sent_ios_per_sec)

        if self.recv_bytes:
            MWorkerLogGen.MWorkerLogAddRecvBytes(builder, self.recv_bytes)

        if self.recv_bytes_per_sec:
            MWorkerLogGen.MWorkerLogAddRecvBytesPerSec(builder, self.recv_bytes_per_sec)

        if self.recv_ios:
            MWorkerLogGen.MWorkerLogAddRecvIos(builder, self.recv_ios)

        if self.recv_ios_per_sec:
            MWorkerLogGen.MWorkerLogAddRecvIosPerSec(builder, self.recv_ios_per_sec)

        if self.router_roles:
            MWorkerLogGen.MWorkerLogAddRouterRoles(builder, self.router_roles)

        if self.router_sessions:
            MWorkerLogGen.MWorkerLogAddRouterSessions(builder, self.router_sessions)

        if self.recv_call:
            MWorkerLogGen.MWorkerLogAddRecvCall(builder, self.recv_call)

        if self.recv_yield:
            MWorkerLogGen.MWorkerLogAddRecvYield(builder, self.recv_yield)

        if self.sent_invocation:
            MWorkerLogGen.MWorkerLogAddSentInvocation(builder, self.sent_invocation)

        if self.sent_result:
            MWorkerLogGen.MWorkerLogAddSentResult(builder, self.sent_result)

        if self.recv_publish:
            MWorkerLogGen.MWorkerLogAddRecvPublish(builder, self.recv_publish)

        if self.sent_published:
            MWorkerLogGen.MWorkerLogAddSentPublished(builder, self.sent_published)

        if self.sent_event:
            MWorkerLogGen.MWorkerLogAddSentEvent(builder, self.sent_event)

        if self.recv_register:
            MWorkerLogGen.MWorkerLogAddRecvRegister(builder, self.recv_register)

        if self.sent_registered:
            MWorkerLogGen.MWorkerLogAddSentRegistered(builder, self.sent_registered)

        if self.recv_subscribe:
            MWorkerLogGen.MWorkerLogAddRecvSubscribe(builder, self.recv_subscribe)

        if self.sent_subscribed:
            MWorkerLogGen.MWorkerLogAddSentSubscribed(builder, self.sent_subscribed)

        final = MWorkerLogGen.MWorkerLogEnd(builder)

        return final



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.xbr.consent import Consents, IndexConsentByMemberAddress
from cfxdb.xbrmm.channel import (
    IndexPayingChannelByDelegate,
    IndexPayingChannelByRecipient,
    IndexPaymentChannelByActor,
    IndexPaymentChannelByDelegate,
    PayingChannelBalances,
    PayingChannels,
    PaymentChannelBalances,
    PaymentChannels,
)
from cfxdb.xbrmm.offer import IndexOfferByKey, Offers
from cfxdb.xbrmm.transaction import Transactions

from .actor import Actors
from .api import Apis, IndexApiByCatalog
from .block import Blocks
from .catalog import Catalogs, IndexCatalogsByOwner
from .market import IndexMarketsByActor, IndexMarketsByMaker, IndexMarketsByOwner, Markets
from .member import Members
from .token import TokenApprovals, TokenTransfers


class Schema(object):
    """
    CFC edge database schema for ZLMDB.
    """

    def __init__(self, db):
        self.db = db

    blocks: Blocks
    """
    Ethereum blocks basic information.
    """

    token_approvals: TokenApprovals
    """
    Token approvals archive.
    """

    token_transfers: TokenTransfers
    """
    Token transfers archive.
    """

    members: Members
    """
    XBR network members.
    """

    catalogs: Catalogs
    """
    XBR network catalogs.
    """

    idx_catalogs_by_owner: IndexCatalogsByOwner
    """
    Index ``(member_oid, created) -> catalog_oid``.
    """

    apis: Apis
    """
    XBR network apis.
    """

    idx_apis_by_catalog: IndexApiByCatalog
    """
    Index ``catalog_oid -> api_oid``.
    """

    markets: Markets
    """
    XBR markets.
    """

    idx_markets_by_owner: IndexMarketsByOwner
    """
    Index ``(owner_adr, created) -> market_oid``.
    """

    idx_markets_by_actor: IndexMarketsByActor
    """
    Index ``(actor_adr, joined) -> market_oid``.
    """

    idx_markets_by_maker: IndexMarketsByMaker
    """
    Index ``maker_adr -> market_oid``.
    """

    actors: Actors
    """
    XBR market actors.
    """

    consents: Consents
    """
    XBR data consents.
    """

    idx_consent_by_member_adr: IndexConsentByMemberAddress
    """
    Consents-by-members-address index with ``(member_adr|bytes[20], joined|int) -> member_adr|UUID`` mapping.
    """

    payment_channels: PaymentChannels
    """
    Payment channels for XBR consumer delegates.
    """

    idx_payment_channel_by_delegate: IndexPaymentChannelByDelegate
    """
    Maps from XBR consumer delegate address to the currently active payment
    channel address for the given consumer delegate.
    """

    idx_payment_channel_by_actor: IndexPaymentChannelByActor
    """
    Maps from XBR (buyer) actor address to the currently active payment
    channel address for the given actor.
    """

    payment_balances: PaymentChannelBalances
    """
    Current off-chain balances within payment channels.
    """

    paying_channels: PayingChannels
    """
    Paying channels for XBR provider delegates.
    """

    idx_paying_channel_by_delegate: IndexPayingChannelByDelegate
    """
    Maps from XBR provider delegate address to the currently active paying
    channel address for the given provider delegate.
    """

    idx_paying_channel_by_recipient: IndexPayingChannelByRecipient
    """
    Maps from XBR seller actor address to the currently active paying
    channel address for the given actor.
    """

    paying_balances: PayingChannelBalances
    """
    Current off-chain balances within paying channels.
    """

    offers: Offers
    """
    Data encryption key offers.
    """

    idx_offer_by_key: IndexOfferByKey
    """
    Index of key offers by key ID (rather than offer ID, as the object table
    is indexed by).
    """

    transactions: Transactions
    """
    """

    @staticmethod
    def attach(db):
        """
        Factory to create a schema from attaching to a database. The schema tables
        will be automatically mapped as persistant maps and attached to the
        database slots.

        :param db: zlmdb.Database
        :return: object of Schema
        """
        schema = Schema(db)

        schema.apis = db.attach_table(Apis)

        schema.idx_apis_by_catalog = db.attach_table(IndexApiByCatalog)

        schema.blocks = db.attach_table(Blocks)

        schema.catalogs = db.attach_table(Catalogs)

        schema.idx_catalogs_by_owner = db.attach_table(IndexCatalogsByOwner)

        schema.catalogs.attach_index(
            "idx1", schema.idx_catalogs_by_owner, lambda catalog: (catalog.owner, catalog.timestamp)
        )

        schema.consents = db.attach_table(Consents)

        schema.idx_consent_by_member_adr = db.attach_table(IndexConsentByMemberAddress)

        schema.token_approvals = db.attach_table(TokenApprovals)

        schema.token_transfers = db.attach_table(TokenTransfers)

        schema.members = db.attach_table(Members)

        schema.markets = db.attach_table(Markets)

        schema.idx_markets_by_owner = db.attach_table(IndexMarketsByOwner)

        schema.markets.attach_index(
            "idx1", schema.idx_markets_by_owner, lambda market: (market.owner, market.timestamp)
        )

        schema.actors = db.attach_table(Actors)
        schema.idx_markets_by_actor = db.attach_table(IndexMarketsByActor)

        # FIXME: we manually maintain the index, and hence must mark the pmap as an index manually
        # schema.idx_markets_by_actor._index_attached_to = schema.actors
        # schema.actors.attach_index('idx1', schema.idx_markets_by_actor, lambda actor:
        #                            (actor.actor, actor.timestamp))

        schema.idx_markets_by_maker = db.attach_table(IndexMarketsByMaker)
        schema.markets.attach_index("idx3", schema.idx_markets_by_maker, lambda market: market.maker)

        schema.payment_channels = db.attach_table(PaymentChannels)

        schema.idx_payment_channel_by_delegate = db.attach_table(IndexPaymentChannelByDelegate)
        schema.payment_channels.attach_index(
            "idx1",
            schema.idx_payment_channel_by_delegate,
            lambda payment_channel: (bytes(payment_channel.delegate), payment_channel.timestamp),
        )

        schema.idx_payment_channel_by_actor = db.attach_table(IndexPaymentChannelByActor)
        schema.payment_channels.attach_index(
            "idx2",
            schema.idx_payment_channel_by_actor,
            lambda payment_channel: (bytes(payment_channel.actor), payment_channel.timestamp),
        )

        schema.payment_balances = db.attach_table(PaymentChannelBalances)

        schema.paying_channels = db.attach_table(PayingChannels)

        schema.idx_paying_channel_by_delegate = db.attach_table(IndexPayingChannelByDelegate)
        schema.paying_channels.attach_index(
            "idx1",
            schema.idx_paying_channel_by_delegate,
            lambda paying_channel: (bytes(paying_channel.delegate), paying_channel.timestamp),
        )

        schema.idx_paying_channel_by_recipient = db.attach_table(IndexPayingChannelByRecipient)
        schema.paying_channels.attach_index(
            "idx2",
            schema.idx_paying_channel_by_recipient,
            lambda paying_channel: (bytes(paying_channel.recipient), paying_channel.timestamp),
        )

        schema.paying_balances = db.attach_table(PayingChannelBalances)

        schema.offers = db.attach_table(Offers)
        schema.idx_offer_by_key = db.attach_table(IndexOfferByKey)
        schema.offers.attach_index("idx1", schema.idx_offer_by_key, lambda offer: offer.key)

        schema.transactions = db.attach_table(Transactions)

        return schema



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

from cfxdb.cookiestore._cookie import Cookie, Cookies, IndexCookiesByValue
from cfxdb.cookiestore._schema import CookieStoreSchema

__all__ = ("Cookie", "Cookies", "IndexCookiesByValue", "CookieStoreSchema")



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: xbrnetwork


class WalletType(object):
    # Unset
    NONE = 0
    # Account (primary) wallet was imported (the user provided the wallet public address).
    IMPORTED = 1
    # Account wallet was configured from Metamask.
    METAMASK = 2
    # Account wallet in hosted (in this database).
    HOSTED = 3



##############################################################################
#
#                        Crossbar.io Database
#     Copyright (c) typedef int GmbH. Licensed under MIT.
#
##############################################################################

import pprint
import uuid
from typing import Any, Dict, Optional

import cbor2
import numpy as np
from zlmdb import MapStringUuid, MapUuidFlatBuffers, flatbuffers, table

from cfxdb.gen.cookiestore import Cookie as CookieGen


class _CookieGen(CookieGen.Cookie):
    @classmethod
    def GetRootAsCookie(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = _CookieGen()
        x.Init(buf, n + offset)
        return x

    def OidAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def AuthenticatedOnNodeAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def AuthenticatedTransportInfoAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None

    def AuthExtraAsBytes(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(36))
        if o != 0:
            _off = self._tab.Vector(o)
            _len = self._tab.VectorLen(o)
            return memoryview(self._tab.Bytes)[_off : _off + _len]
        return None


class Cookie(object):
    """
    Persistent cookies, as used in WAMP-Cookie authentication by router and proxy workers.
    """

    def __init__(self, from_fbs=None):
        self._from_fbs = from_fbs

        # [uint8] (uuid)
        self._oid = None

        # uint64 (timestamp)
        self._created = None

        # uint64
        self._max_age = None

        # string
        self._name = None

        # string
        self._value = None

        # uint64 (timestamp)
        self._authenticated = None

        # [uint8] (uuid)
        self._authenticated_on_node = None

        # string
        self._authenticated_on_worker = None

        # [uint8] (cbor)
        self._authenticated_transport_info = None

        # uint64
        self._authenticated_session = None

        # uint64 (timestamp)
        self._authenticated_joined_at = None

        # string
        self._authenticated_authmethod = None

        # string
        self._authid = None

        # string
        self._authrole = None

        # string
        self._authmethod = None

        # string
        self._authrealm = None

        # [uint8] (cbor)
        self._authextra = None

    def marshal(self) -> dict:
        obj = {
            "oid": self.oid.bytes if self.oid else None,
            "created": int(self.created) if self.created else None,
            "max_age": self.max_age,
            "name": self.name,
            "value": self.value,
            "authenticated": int(self.authenticated) if self.authenticated else None,
            "authenticated_on_node": self.authenticated_on_node.bytes if self.authenticated_on_node else None,
            "authenticated_on_worker": self.authenticated_on_worker,
            "authenticated_transport_info": self.authenticated_transport_info,
            "authenticated_session": self.authenticated_session,
            "authenticated_joined_at": int(self.authenticated_joined_at) if self.authenticated_joined_at else None,
            "authenticated_authmethod": self.authenticated_authmethod,
            "authid": self.authid,
            "authrole": self.authrole,
            "authrealm": self.authrealm,
            "authextra": self.authextra,
        }
        return obj

    def __str__(self):
        return "\n{}\n".format(pprint.pformat(self.marshal()))

    @property
    def oid(self) -> uuid.UUID:
        """
        Database ID of this cookie record.
        """
        if self._oid is None and self._from_fbs:
            if self._from_fbs.OidLength():
                _oid = self._from_fbs.OidAsBytes()
                self._oid = uuid.UUID(bytes=bytes(_oid))
        return self._oid

    @oid.setter
    def oid(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._oid = value

    @property
    def created(self) -> np.datetime64:
        """
        Timestamp when the cookie was created. Epoch time in ns.
        """
        if self._created is None and self._from_fbs:
            self._created = np.datetime64(self._from_fbs.Created(), "ns")
        return self._created

    @created.setter
    def created(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._created = value

    @property
    def max_age(self) -> int:
        """
        Cookie maximum age (lifetime of the cookie in seconds, see http://tools.ietf.org/html/rfc6265#page-20),
            e.g. ``604800"``.
        """
        if self._max_age is None and self._from_fbs:
            self._max_age = self._from_fbs.MaxAge()
        return self._max_age

    @max_age.setter
    def max_age(self, value: int):
        assert value is None or type(value) == int
        self._max_age = value

    @property
    def name(self) -> str:
        """
        Cookie name, as set in HTTP header, e.g. ``"cbtid"``.
        """
        if self._name is None and self._from_fbs:
            _name = self._from_fbs.Name()
            if _name:
                self._name = _name.decode("utf8")
        return self._name

    @name.setter
    def name(self, value: str):
        self._name = value

    @property
    def value(self) -> str:
        """
        Cookie value, as set in HTTP header, e.g. ``"gn2ri8fuAYQse50/L6N7jnt2"``.
        """
        if self._value is None and self._from_fbs:
            _value = self._from_fbs.Value()
            if _value:
                self._value = _value.decode("utf8")
        return self._value

    @value.setter
    def value(self, value: str):
        self._value = value

    @property
    def authenticated(self) -> np.datetime64:
        """
        Timestamp when the cookie was authenticated (if any). Epoch time in ns.
        """
        if self._authenticated is None and self._from_fbs:
            self._authenticated = np.datetime64(self._from_fbs.Authenticated(), "ns")
        return self._authenticated

    @authenticated.setter
    def authenticated(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._authenticated = value

    @property
    def authenticated_on_node(self) -> uuid.UUID:
        """
        The Crossbar.io node (within the management domain) the cookie was authenticated on (if any).
        """
        if self._authenticated_on_node is None and self._from_fbs:
            if self._from_fbs.AuthenticatedOnNodeLength():
                _authenticated_on_node = self._from_fbs.AuthenticatedOnNodeAsBytes()
                self._authenticated_on_node = uuid.UUID(bytes=bytes(_authenticated_on_node))
        return self._authenticated_on_node

    @authenticated_on_node.setter
    def authenticated_on_node(self, value: uuid.UUID):
        assert value is None or isinstance(value, uuid.UUID)
        self._authenticated_on_node = value

    @property
    def authenticated_on_worker(self) -> str:
        """
        The Crossbar.io worker (within the node) the cookie was authenticated on (if any).
        """
        if self._authenticated_on_worker is None and self._from_fbs:
            _authenticated_on_worker = self._from_fbs.AuthenticatedOnWorker()
            if _authenticated_on_worker:
                self._authenticated_on_worker = _authenticated_on_worker.decode("utf8")
        return self._authenticated_on_worker

    @authenticated_on_worker.setter
    def authenticated_on_worker(self, value: str):
        self._authenticated_on_worker = value

    @property
    def authenticated_transport_info(self) -> dict:
        """
        The client transport information for the connection the cookie was authenticated in (if any).
        """
        if self._authenticated_transport_info is None and self._from_fbs:
            _authenticated_transport_info = self._from_fbs.AuthenticatedTransportInfoAsBytes()
            if _authenticated_transport_info:
                self._authenticated_transport_info = cbor2.loads(_authenticated_transport_info)
            else:
                self._authenticated_transport_info = {}
        return self._authenticated_transport_info

    @authenticated_transport_info.setter
    def authenticated_transport_info(self, value: Optional[Dict[str, Any]]):
        assert value is None or type(value) == dict
        self._authenticated_transport_info = value

    @property
    def authenticated_session(self) -> int:
        """
        The WAMP session ID of the original authenticating session.
        """
        if self._authenticated_session is None and self._from_fbs:
            self._authenticated_session = self._from_fbs.AuthenticatedSession()
        return self._authenticated_session

    @authenticated_session.setter
    def authenticated_session(self, value: int):
        assert value is None or type(value) == int
        self._authenticated_session = value

    @property
    def authenticated_joined_at(self) -> np.datetime64:
        """
        Timestamp when the original authenticating session was welcome by the router. Epoch time in ns.
        """
        if self._authenticated_joined_at is None and self._from_fbs:
            self._authenticated_joined_at = np.datetime64(self._from_fbs.AuthenticatedJoinedAt(), "ns")
        return self._authenticated_joined_at

    @authenticated_joined_at.setter
    def authenticated_joined_at(self, value: np.datetime64):
        assert value is None or isinstance(value, np.datetime64)
        self._authenticated_joined_at = value

    @property
    def authenticated_authmethod(self) -> str:
        """
        The (original) WAMP authentication method, after which the client was authenticated before setting this cookie.
        """
        if self._authenticated_authmethod is None and self._from_fbs:
            _authenticated_authmethod = self._from_fbs.AuthenticatedAuthmethod()
            if _authenticated_authmethod:
                self._authenticated_authmethod = _authenticated_authmethod.decode("utf8")
        return self._authenticated_authmethod

    @authenticated_authmethod.setter
    def authenticated_authmethod(self, value: str):
        self._authenticated_authmethod = value

    @property
    def authid(self) -> str:
        """
        The WAMP authid a cookie-authenticating session is to be assigned.
        """
        if self._authid is None and self._from_fbs:
            _authid = self._from_fbs.Authid()
            if _authid:
                self._authid = _authid.decode("utf8")
        return self._authid

    @authid.setter
    def authid(self, value: str):
        self._authid = value

    @property
    def authrole(self) -> str:
        """
        The WAMP authrole a cookie-authenticating session is to join under.
        """
        if self._authrole is None and self._from_fbs:
            _authrole = self._from_fbs.Authrole()
            if _authrole:
                self._authrole = _authrole.decode("utf8")
        return self._authrole

    @authrole.setter
    def authrole(self, value: str):
        self._authrole = value

    @property
    def authmethod(self) -> str:
        """
        The WAMP authmethod a cookie-authenticating session is to join under.
        """
        if self._authmethod is None and self._from_fbs:
            _authmethod = self._from_fbs.Authmethod()
            if _authmethod:
                self._authmethod = _authmethod.decode("utf8")
        return self._authmethod

    @authmethod.setter
    def authmethod(self, value: str):
        self._authmethod = value

    @property
    def authrealm(self) -> str:
        """
        The WAMP realm a cookie-authenticating session is to join.
        """
        if self._authrealm is None and self._from_fbs:
            _authrealm = self._from_fbs.Authrealm()
            if _authrealm:
                self._authrealm = _authrealm.decode("utf8")
        return self._authrealm

    @authrealm.setter
    def authrealm(self, value: str):
        self._authrealm = value

    @property
    def authextra(self) -> dict:
        """
        The WAMP authentication extra data to be returned to the client performing cookie-based authentication.
        """
        if self._authextra is None and self._from_fbs:
            _authextra = self._from_fbs.AuthExtraAsBytes()
            if _authextra:
                self._authextra = cbor2.loads(_authextra)
            else:
                self._authextra = {}
        return self._authextra

    @authextra.setter
    def authextra(self, value: Optional[Dict[str, Any]]):
        assert value is None or type(value) == dict
        self._authextra = value

    @staticmethod
    def cast(buf) -> "Cookie":
        return Cookie(_CookieGen.GetRootAsCookie(buf, 0))

    def build(self, builder):
        oid = self.oid.bytes if self.oid else None
        if oid:
            oid = builder.CreateString(oid)

        name = self.name
        if name:
            name = builder.CreateString(name)

        value = self.value
        if value:
            value = builder.CreateString(value)

        authenticated_on_node = self.authenticated_on_node.bytes if self.authenticated_on_node else None
        if authenticated_on_node:
            authenticated_on_node = builder.CreateString(authenticated_on_node)

        authenticated_on_worker = self.authenticated_on_worker
        if authenticated_on_worker:
            authenticated_on_worker = builder.CreateString(authenticated_on_worker)

        authenticated_transport_info = self.authenticated_transport_info
        if authenticated_transport_info:
            authenticated_transport_info = builder.CreateString(cbor2.dumps(authenticated_transport_info))

        authenticated_authmethod = self.authenticated_authmethod
        if authenticated_authmethod:
            authenticated_authmethod = builder.CreateString(authenticated_authmethod)

        authid = self.authid
        if authid:
            authid = builder.CreateString(authid)

        authrole = self.authrole
        if authrole:
            authrole = builder.CreateString(authrole)

        authmethod = self.authmethod
        if authmethod:
            authmethod = builder.CreateString(authmethod)

        authrealm = self.authrealm
        if authrealm:
            authrealm = builder.CreateString(authrealm)

        authextra = self.authextra
        if authextra:
            authextra = builder.CreateString(cbor2.dumps(authextra))

        CookieGen.CookieStart(builder)

        if oid:
            CookieGen.CookieAddOid(builder, oid)

        if self.created:
            CookieGen.CookieAddCreated(builder, int(self.created))

        if self.max_age:
            CookieGen.CookieAddMaxAge(builder, self.max_age)

        if name:
            CookieGen.CookieAddName(builder, name)

        if value:
            CookieGen.CookieAddValue(builder, value)

        if self.authenticated:
            CookieGen.CookieAddAuthenticated(builder, int(self.authenticated))

        if authenticated_on_node:
            CookieGen.CookieAddAuthenticatedOnNode(builder, authenticated_on_node)

        if authenticated_on_worker:
            CookieGen.CookieAddAuthenticatedOnWorker(builder, authenticated_on_worker)

        if authenticated_transport_info:
            CookieGen.CookieAddAuthenticatedTransportInfo(builder, authenticated_transport_info)

        if self.authenticated_session:
            CookieGen.CookieAddAuthenticatedSession(builder, self.authenticated_session)

        if self.authenticated_joined_at:
            CookieGen.CookieAddAuthenticatedJoinedAt(builder, int(self.authenticated_joined_at))

        if authenticated_authmethod:
            CookieGen.CookieAddAuthenticatedAuthmethod(builder, authenticated_authmethod)

        if authid:
            CookieGen.CookieAddAuthid(builder, authid)

        if authrole:
            CookieGen.CookieAddAuthrole(builder, authrole)

        if authmethod:
            CookieGen.CookieAddAuthmethod(builder, authmethod)

        if authrealm:
            CookieGen.CookieAddAuthrealm(builder, authrealm)

        if authextra:
            CookieGen.CookieAddAuthextra(builder, authextra)

        final = CookieGen.CookieEnd(builder)

        return final


@table("62f8c8c9-c50b-4686-bafe-38b221c64a0c", build=Cookie.build, cast=Cookie.cast)
class Cookies(MapUuidFlatBuffers):
    """
    Persisted cookies table.

    Map :class:`zlmdb.MapUuidFlatBuffers` from ``cookie_oid`` to :class:`cfxdb.cookiestore.Cookie`
    """


@table("65e1d8c1-fa8b-459d-ae43-cb320d28cc97")
class IndexCookiesByValue(MapStringUuid):
    """
    Index: cookie_value -> cookie_oid
    """



# Register zlmdb's vendored flatbuffers in sys.modules so that the flatc-generated
# code in this package (which does `import flatbuffers`) resolves to zlmdb's
# vendored copy instead of requiring a separate flatbuffers package.
#
# Note: We import zlmdb directly here (not via cfxdb) to avoid circular imports.
# The setup_flatbuffers_import function may be monkey-patched by cfxdb.__init__
# for zlmdb < 25.12.2, or natively available in zlmdb >= 25.12.2.
import sys
import zlmdb

# Check if zlmdb has setup_flatbuffers_import (zlmdb >= 25.12.2 or monkey-patched)
if hasattr(zlmdb, "setup_flatbuffers_import"):
    zlmdb.setup_flatbuffers_import()
else:
    # Fallback for zlmdb 25.12.1: register vendored flatbuffers directly
    _vendor = zlmdb._flatbuffers_vendor
    sys.modules.setdefault("flatbuffers", _vendor)
    sys.modules.setdefault("flatbuffers.compat", _vendor.compat)
    sys.modules.setdefault("flatbuffers.builder", _vendor.builder)
    sys.modules.setdefault("flatbuffers.table", _vendor.table)
    sys.modules.setdefault("flatbuffers.util", _vendor.util)
    sys.modules.setdefault("flatbuffers.number_types", _vendor.number_types)
    sys.modules.setdefault("flatbuffers.packer", _vendor.packer)
    sys.modules.setdefault("flatbuffers.encode", _vendor.encode)
